# SPDX-License-Identifier: AGPL-3.0-or-later
"""Forms for the names app."""

import re

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.modules import names
from plinth.utils import format_lazy

HOSTNAME_REGEX = r'^[a-zA-Z0-9]([-a-zA-Z0-9]{,61}[a-zA-Z0-9])?$'


class NamesConfigurationForm(forms.Form):
    """Form to configure names app."""

    dns_over_tls = forms.ChoiceField(
        label=_('Use DNS-over-TLS for resolving domains (global preference)'),
        widget=forms.RadioSelect, choices=[
            ('yes',
             format_lazy(
                 'Yes. Encrypt connections to the DNS server. <p '
                 'class="help-block">This improves privacy as domain name '
                 'queries will not be made as plain text over the network. It '
                 'also improves security as responses from the server cannot '
                 'be manipulated. If the configured DNS servers do not '
                 'support DNS-over-TLS, all name resolutions will fail. If '
                 'your DNS provider (likely your ISP) does not support '
                 'DNS-over-TLS or blocks some domains, you can configure '
                 'well-known public DNS servers in individual network '
                 'connection settings.</p>', allow_markup=True)),
            ('opportunistic',
             format_lazy(
                 'Opportunistic. <p class="help-block">Encrypt connections to '
                 'the DNS server if the server supports DNS-over-TLS. '
                 'Otherwise, use unencrypted connections. There is no '
                 'protection against response manipulation.</p>',
                 allow_markup=True)),
            ('no',
             format_lazy(
                 'No. <p class="help-block">Do not encrypt domain name '
                 'resolutions.</p>', allow_markup=True)),
        ], initial='no')

    dnssec = forms.ChoiceField(
        label=_('Use DNSSEC when resolving domains (global preference)'),
        widget=forms.RadioSelect, choices=[
            ('yes',
             format_lazy(
                 'Yes. Verify authenticity and integrity of domain '
                 'resolutions. <p class="help-block">This improves security. '
                 'If the configured DNS servers do not support DNSSEC, all '
                 'name resolutions will fail. If your DNS provider (likely '
                 'your ISP) does not support DNSSEC or is manipulating '
                 'responses, you can configure well-known public DNS servers '
                 'in individual network connection settings.</p>',
                 allow_markup=True)),
            ('allow-downgrade',
             format_lazy(
                 'Allow downgrade. <p class="help-block">Verify name '
                 'resolutions done by the DNS server if the server supports '
                 'DNSSEC. Otherwise, allow unverified resolutions. Limited '
                 'improvement to security. Detecting whether a DNS server '
                 'supports DNSSEC is not very reliable currently.</p>',
                 allow_markup=True)),
            ('no',
             format_lazy(
                 'No. <p class="help-block">Do not verify domain name '
                 'resolutions.</p>', allow_markup=True)),
        ], initial='no')


class HostnameForm(forms.Form):
    """Form to update system's hostname."""
    # See:
    # https://tools.ietf.org/html/rfc952
    # https://tools.ietf.org/html/rfc1035#section-2.3.1
    # https://tools.ietf.org/html/rfc1123#section-2
    # https://tools.ietf.org/html/rfc2181#section-11
    hostname = forms.CharField(
        label=_('Hostname'), help_text=format_lazy(
            _('Hostname is the local name by which other devices on the local '
              'network can reach your {box_name}.  It must start and end with '
              'an alphabet or a digit and have as interior characters only '
              'alphabets, digits and hyphens.  Total length must be 63 '
              'characters or less.'), box_name=_(cfg.box_name)), validators=[
                  validators.RegexValidator(HOSTNAME_REGEX,
                                            _('Invalid hostname'))
              ], strip=True)


def _domain_label_validator(domain_name):
    """Validate domain name labels."""
    for label in domain_name.split('.'):
        if not re.match(HOSTNAME_REGEX, label):
            raise ValidationError(_('Invalid domain name'))


class DomainAddForm(forms.Form):
    """Form to add a static domain name."""

    domain_name = forms.CharField(
        label=_('Domain Name'), help_text=format_lazy(
            _('Domain name is the global name by which other devices on the '
              'Internet can reach your {box_name}.  It must consist of '
              'labels separated by dots.  Each label must start and end '
              'with an alphabet or a digit and have as interior characters '
              'only alphabets, digits and hyphens.  Length of each label '
              'must be 63 characters or less.  Total length of domain name '
              'must be 253 characters or less.'), box_name=_(cfg.box_name)),
        required=True, validators=[
            validators.RegexValidator(
                r'^[a-zA-Z0-9]([-a-zA-Z0-9.]{,251}[a-zA-Z0-9])?$',
                _('Invalid domain name')), _domain_label_validator
        ], strip=True)

    def clean_domain_name(self):
        """Check if the name is valid."""
        domain_name = self.cleaned_data['domain_name']
        if domain_name in names.domains_list():
            raise ValidationError(_('Domain already exists.'))

        return domain_name



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for custom Django template tags.
"""

from unittest.mock import patch

from plinth.templatetags import extras


def _assert_active_url(menu, url):
    """Verify that only the given url is set as 'active' in the menu"""
    for item in menu:
        if item['url'] == url:
            assert item['active']
        else:
            assert not item['active']


def _verify_active_menuitems(menu):
    """Verify that one and only one menuitem is marked as active"""
    return sum([item['active'] for item in menu]) == 1


def test_highlighting():
    """Test detection of active subsubmenu items using request.path"""
    menu = [{
        'url': '/abc/123/abc/',
        'text': 'abc'
    }, {
        'url': '/abc/123/',
        'text': 'overview'
    }, {
        'url': '/abc/123/crunch/',
        'text': 'crunch'
    }, {
        'url': '/abc/123/create/',
        'text': 'create'
    }]

    tests = [['/abc/123/crunch/new/', '/abc/123/crunch/'],
             ['/abc/123/create/', '/abc/123/create/'],
             ['/abc/123/nolink/', '/abc/123/'], ['/abc/123/abx/', '/abc/123/'],
             ['/abc/123/ab/', '/abc/123/'], ['/abc/123/', '/abc/123/']]

    for check_path, expected_active_path in tests:
        menu = extras.mark_active_menuitem(menu, check_path)
        _assert_active_url(menu, expected_active_path)
        assert _verify_active_menuitems(menu)


@patch('plinth.web_server.resolve_static_path')
def test_icon(resolve_static_path, tmp_path):
    """Test that the icon tag works for basic usage."""
    icon1 = tmp_path / 'icon1.svg'
    icon1.write_text('<svg><cirlcle></circle></svg>')
    resolve_static_path.return_value = icon1
    return_value = extras.icon('icon1')
    assert return_value == ('<svg class="svg-icon" data-icon-name="icon1" '
                            '><cirlcle></circle></svg>')


@patch('plinth.web_server.resolve_static_path')
def test_icon_attributes(resolve_static_path, tmp_path):
    """Test that the icon tag works with attributes."""
    icon1 = tmp_path / 'icon1.svg'
    icon1.write_text('<svg><cirlcle></circle></svg>')
    resolve_static_path.return_value = icon1

    attributes = {'class': 'test-class'}
    return_value = extras.icon('icon1', **attributes)
    assert return_value == ('<svg class="test-class" data-icon-name="icon1" '
                            '><cirlcle></circle></svg>')

    attributes = {'data-test': 'test-value'}
    return_value = extras.icon('icon1', **attributes)
    assert return_value == ('<svg data-test="test-value" class="svg-icon" '
                            'data-icon-name="icon1" ><cirlcle></circle></svg>')


@patch('plinth.utils.random_string')
@patch('plinth.web_server.resolve_static_path')
def test_icon_auto_ids(resolve_static_path, random_string, tmp_path):
    """Test that the icon tag works for implementing automatic IDs."""
    random_string.return_value = 'randomvalue'
    icon2 = tmp_path / 'icon2.svg'
    icon2.write_text('<svg><cirlcle id="autoidmagic-foo"></circle>'
                     '<path id="autoidmagic-bar"></path></svg>')
    resolve_static_path.return_value = icon2
    return_value = extras.icon('icon2')
    assert return_value == ('<svg class="svg-icon" data-icon-name="icon2" >'
                            '<cirlcle id="randomvalue-foo"></circle>'
                            '<path id="randomvalue-bar"></path></svg>')


@patch('plinth.web_server.resolve_static_path')
def test_icon_xml_stripping(resolve_static_path, tmp_path):
    """Test that the icon tag strips the XML header."""
    icon2 = tmp_path / 'icon2.svg'
    icon2.write_text('<?xml version="1.0" encoding="utf-8">\n'
                     '<svg><cirlcle></circle></svg>')
    resolve_static_path.return_value = icon2
    return_value = extras.icon('icon2')
    assert return_value == ('<svg class="svg-icon" data-icon-name="icon2" >'
                            '<cirlcle></circle></svg>')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Module to interact with systemd-resolved over DBus."""

import ipaddress
import json
import subprocess

from django.utils.translation import gettext_lazy as _

from plinth.utils import import_from_gi

gio = import_from_gi('Gio', '2.0')

RESOLVE_NAME = 'org.freedesktop.resolve1'
RESOLVE_PATH = '/org/freedesktop/resolve1'
MANAGER_INTERFACE = 'org.freedesktop.resolve1.Manager'
LINK_INTERFACE = 'org.freedesktop.resolve1.Link'


class DNSServer:
    """Representation of a DNS server in systemd-resolved state."""

    def __init__(self, link_index: int, address_class: int,
                 address_ints: list[int], port: int = 0,
                 domain_name: str | None = None):
        self.link_index = link_index
        self.address_class = address_class
        self.address = ipaddress.ip_address(bytes(address_ints))
        self.port = port
        self.domain_name = domain_name

    def __str__(self):
        if self.port:
            if self.address.version == 4:  # IPv4
                address_str = f'{self.address.compressed}:{self.port}'
            else:  # IPv6
                address_str = f'[{self.address.compressed}]:{self.port}'
        else:
            address_str = self.address.compressed

        if self.domain_name:
            return f'{address_str} ({self.domain_name})'

        return address_str


class Link:
    """systemd-resolved state for a particular link or global context."""

    def __init__(self, connection, object_path, link_index: int = 0,
                 interface_name: str | None = None):
        """Fetch all the relevant properties for a link over DBus."""
        if not link_index:  # Global
            interface = MANAGER_INTERFACE
        else:
            interface = LINK_INTERFACE

        self.proxy = gio.DBusProxy.new_sync(connection,
                                            gio.DBusProxyFlags.NONE, None,
                                            RESOLVE_NAME, object_path,
                                            interface)

        self.link_index = link_index
        self.interface_name = interface_name

        self.dns_over_tls = self.proxy.get_cached_property(
            'DNSOverTLS').unpack()
        self.dnssec = self.proxy.get_cached_property('DNSSEC').unpack()
        self.dnssec_supported = self.proxy.get_cached_property(
            'DNSSECSupported')

        self.dns_servers = self._new_dns_servers(
            self.proxy.get_cached_property('DNSEx'))

        self.fallback_dns_servers = None
        if not link_index:
            self.fallback_dns_servers = self._new_dns_servers(
                self.proxy.get_cached_property('FallbackDNSEx'))

        self.current_dns_server = self._new_dns_server(
            self.proxy.get_cached_property('CurrentDNSServerEx'))

    def get_link(self, link_index):
        """Return a string path to a link's DBus object."""
        return self.proxy.GetLink('(i)', link_index)

    @property
    def dns_over_tls_string(self):
        """Return a string representation for DNS-over-TLS status."""
        value_map = {
            'yes': _('yes'),
            'opportunistic': _('opportunistic'),
            'no': _('no')
        }
        return value_map.get(self.dns_over_tls, self.dns_over_tls)

    @property
    def dnssec_string(self):
        """Return a string representation for DNSSEC status."""
        value_map = {
            'yes': _('yes'),
            'allow-downgrade': _('allow-downgrade'),
            'no': _('no')
        }
        return value_map.get(self.dnssec, self.dnssec)

    @property
    def dnssec_supported_string(self):
        """Return a string representation for whether DNSSEC is supported."""
        return _('supported') if self.dnssec_supported else _('unsupported')

    def _new_dns_servers(self, dns_tuples):
        """Return list of DNS Server objects from variant tuple.

        Global DNS servers list also contains individual link DNS servers.
        Ignore those.
        """
        return [
            self._new_dns_server(dns_tuple) for dns_tuple in dns_tuples
            if self.link_index != 0 or dns_tuple[0] == 0
        ]

    def _new_dns_server(self, dns_tuple):
        """Return a DNS Server object from variant tuple.

        Tuple can be prefixed by link index in case of DNS server for global
        context. Handle both cases. Entire tuple may be empty. Return None in
        that case.
        """
        if self.link_index:  # Not global
            dns_tuple = (self.link_index, ) + tuple(dns_tuple)

        if not dns_tuple[2]:  # Empty address
            return None

        return DNSServer(*dns_tuple)

    def __str__(self):
        dnssec_supported = ('supported'
                            if self.dnssec_supported else 'unspported')
        value = ''
        if not self.link_index:
            value += 'Global\n'
        else:
            value = f'Link {self.link_index} ({self.interface_name})\n'

        if self.current_dns_server:
            value += f'  Current DNS Server: {str(self.current_dns_server)}\n'

        if self.dns_servers:
            value += '  DNS Servers:\n'
            for server in self.dns_servers:
                value += f'    {server}\n'

        if self.fallback_dns_servers:
            value += '  Fallback DNS Servers: \n'
            for server in self.fallback_dns_servers:
                value += f'    {server}\n'

        value += f'  DNS-over-TLS: {self.dns_over_tls}\n'
        value += f'  DNSSEC: {self.dnssec}/{dnssec_supported}\n'
        return value


def get_links():
    """Return the list of network interfaces and their indices."""
    process = subprocess.run(['ip', '--json', 'link'], stdout=subprocess.PIPE,
                             check=True)
    output = json.loads(process.stdout.decode())

    links = {}  # Maintain link index order
    for entry in output:
        links[entry['ifindex']] = entry['ifname']

    return links


def get_status():
    """Return the current status of systemd-resolved daemon."""
    link_status = []
    connection = gio.bus_get_sync(gio.BusType.SYSTEM)
    global_link = Link(connection, RESOLVE_PATH)
    link_status.append(global_link)

    links = get_links()
    for link_index, interface_name in links.items():
        if interface_name == 'lo':
            continue

        link_path = global_link.get_link(link_index)
        link_status.append(
            Link(connection, link_path, link_index, interface_name))

    return link_status



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Privileged actions for Kiwix content server.
"""

import pathlib
import shutil
import subprocess
from xml.etree import ElementTree

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules import kiwix

# Only one central library is supported.
KIWIX_HOME = pathlib.Path('/var/lib/kiwix-server-freedombox')
LIBRARY_FILE = KIWIX_HOME / 'library_zim.xml'
CONTENT_DIR = KIWIX_HOME / 'content'


@privileged
def add_package(file_name: str, temporary_file_path: str):
    """Adds a content package to Kiwix.

    Adding packages is idempotent.

    Users can add content to Kiwix in multiple ways:
     - Upload a ZIM file
     - Provide a link to the ZIM file
     - Provide a magnet link to the ZIM file

    The commandline download manager aria2c is a dependency of kiwix-tools.
    aria2c is used for both HTTP and Magnet downloads.
    """
    kiwix.validate_file_name(file_name)
    # file_name is verified further by move_uploaded_file()

    # Moving files to the Kiwix library path ensures that
    # they can't be removed by other apps or users.
    CONTENT_DIR.mkdir(exist_ok=True)
    action_utils.move_uploaded_file(temporary_file_path, CONTENT_DIR,
                                    file_name, allow_overwrite=False,
                                    user='root', group='root',
                                    permissions=0o644)

    content_file = CONTENT_DIR / file_name
    try:
        _kiwix_manage_add(str(CONTENT_DIR / file_name))
    except subprocess.CalledProcessError:
        # Remove the uploaded file if we could not added it to library
        content_file.unlink()
        raise


def _kiwix_manage_add(zim_file: str):
    action_utils.run(['kiwix-manage', LIBRARY_FILE, 'add', zim_file],
                     check=True)

    # kiwix-serve doesn't read the library file unless it is restarted.
    action_utils.service_try_restart('kiwix-server-freedombox')


@privileged
def uninstall() -> None:
    """Remove all content during uninstall."""
    shutil.rmtree(str(CONTENT_DIR), ignore_errors=True)
    LIBRARY_FILE.unlink(missing_ok=True)


@privileged
def list_packages() -> dict[str, dict[str, str]]:
    """Return the list of content packages configured in library file."""
    library = ElementTree.parse(LIBRARY_FILE).getroot()

    books = {}
    for book in library:
        path = book.attrib['path'].split('/')[-1]
        path = path.removesuffix('.zim').lower()  # Strip '.zim' from the path
        try:
            books[book.attrib['id']] = {
                'title': book.attrib['title'],
                'description': book.attrib['description'],
                'path': path
            }
        except KeyError:
            pass  # Ignore entries that don't have expected properties

    return books


@privileged
def delete_package(zim_id: str):
    """Remove a content package from the library file."""
    library = ElementTree.parse(LIBRARY_FILE).getroot()

    for book in library:
        try:
            if book.attrib['id'] != zim_id:
                continue

            action_utils.run(['kiwix-manage', LIBRARY_FILE, 'remove', zim_id],
                             check=True)
            (KIWIX_HOME / book.attrib['path']).unlink()
            action_utils.service_try_restart('kiwix-server-freedombox')
            return
        except KeyError:  # Expected properties not found on elements
            pass



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities for managing WireGuard."""

import datetime
import logging
import subprocess
import time

from plinth import app as app_module
from plinth import network
from plinth.utils import import_from_gi

from . import privileged

nm = import_from_gi('NM', '1.0')

IP_TEMPLATE = '10.84.0.{}'

logger = logging.getLogger(__name__)


def _get_nm_address_info(settings_ipv4, settings_ipv6) -> tuple[str, str]:
    """Extract IP address info from NM IPv4/IPv6 settings."""
    for settings in [settings_ipv4, settings_ipv6]:
        if settings and settings.get_num_addresses():
            nm_address = settings.get_address(0)
            address = nm_address.get_address()
            prefix = str(nm_address.get_prefix())
            return address, address + '/' + prefix

    return '', ''


def get_nm_info():
    """Get information from network manager."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    client = network.get_nm_client()

    connections = {}
    for connection in client.get_connections():
        if connection.get_connection_type() != setting_name:
            continue

        settings = connection.get_setting_by_name(setting_name)
        secrets = connection.get_secrets(setting_name)
        connection.update_secrets(setting_name, secrets)

        info = {}
        info['interface'] = connection.get_interface_name()
        info['private_key'] = settings.get_private_key()
        info['public_key'] = None
        info['listen_port'] = settings.get_listen_port()
        info['fwmark'] = settings.get_fwmark()
        info['mtu'] = settings.get_mtu()
        info['peers'] = {}
        for peer_index in range(settings.get_peers_len()):
            peer = settings.get_peer(peer_index)
            peer_info = {
                'endpoint': peer.get_endpoint(),
                'public_key': peer.get_public_key(),
                'preshared_key': peer.get_preshared_key(),
                'persistent_keepalive': peer.get_persistent_keepalive(),
                'allowed_ips': []
            }
            for index in range(peer.get_allowed_ips_len()):
                allowed_ip = peer.get_allowed_ip(index, None)
                peer_info['allowed_ips'].append(allowed_ip)

            info['peers'][peer_info['public_key']] = peer_info

        # This is default route when all IPs are in allowed IPs list.
        info['default_route'] = False
        for peer_info in info['peers'].values():
            if ('0.0.0.0/0' in peer_info['allowed_ips']
                    and '::/0' in peer_info['allowed_ips']):
                info['default_route'] = True

        settings_ipv4 = connection.get_setting_ip4_config()
        settings_ipv6 = connection.get_setting_ip6_config()

        ip_address, ip_address_and_network = _get_nm_address_info(
                settings_ipv4, settings_ipv6
        )

        info['ip_address'] = ip_address
        info['ip_address_and_network'] = ip_address_and_network

        connections[info['interface']] = info

    return connections


def get_info():
    """Return server and clients info."""
    status = privileged.get_info()

    nm_info = get_nm_info()

    my_server_info = None
    my_client_servers = {}
    for interface, info in nm_info.items():
        if interface == 'wg0':
            my_server_info = info
        else:
            my_client_servers[interface] = info

        # If the NM connection is not active but the device link is up, 'wg
        # show' will not show any public key configured on the interface.
        if interface not in status or (interface in status and
                                       not status[interface]['public_key']):
            info['public_key'] = _get_public_key_from_private_key(
                info['private_key'])
            continue

        info['public_key'] = status[interface]['public_key']
        for status_peer in status[interface]['peers']:
            if status_peer['latest_handshake']:
                status_peer['latest_handshake'] = \
                    datetime.datetime.fromtimestamp(
                        status_peer['latest_handshake'])
            public_key = status_peer['public_key']
            info_peer = info['peers'].setdefault(public_key, {})
            info_peer['status'] = status_peer

    return {
        'my_server': my_server_info,
        'my_client': {
            'servers': my_client_servers,
        },
    }


def enable_connections(enable):
    """Activate all connections and set them to auto-connect."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    client = network.get_nm_client()
    for connection in client.get_connections():
        if connection.get_connection_type() != setting_name:
            continue

        network.edit_connection(connection,
                                {'common': {
                                    'autoconnect': enable
                                }})
        if enable:
            network.activate_connection(connection.get_uuid())
        else:
            try:
                network.deactivate_connection(connection.get_uuid())
            except network.ConnectionNotFound:
                pass  # Connection is already inactive


def delete_connections():
    """Remove all WireGuard connections."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    client = network.get_nm_client()
    for connection in client.get_connections():
        if connection.get_connection_type() != setting_name:
            continue

        connection.delete()


def _get_public_key_from_private_key(private_key):
    process = subprocess.run(['wg', 'pubkey'], check=True, capture_output=True,
                             input=private_key.encode())
    return process.stdout.decode()


def _generate_private_key():
    """Return a private key generated by 'wg' command."""
    process = subprocess.run(['wg', 'genkey'], check=True, capture_output=True)
    return process.stdout.decode().strip()


def generate_client_keypair():
    """Generate client private/public keypair."""
    private_key = _generate_private_key()
    public_key = _get_public_key_from_private_key(private_key)
    return private_key, public_key


def _find_next_interface():
    """Find next unused wireguard interface name."""
    output = subprocess.check_output(['wg', 'show',
                                      'interfaces']).decode().strip()
    interfaces = output.split()
    interface_num = 1
    new_interface_name = 'wg1'
    while new_interface_name in interfaces:
        interface_num += 1
        new_interface_name = 'wg' + str(interface_num)

    return new_interface_name


def add_server(settings):
    """Add information for connecting to a server."""
    app = app_module.App.get('wireguard')
    interface_name = _find_next_interface()
    settings['common']['name'] = 'WireGuard-Client-' + interface_name
    settings['common']['interface'] = interface_name
    settings['common']['autoconnect'] = app.is_enabled()
    if not settings['wireguard']['private_key']:
        settings['wireguard']['private_key'] = _generate_private_key()

    network.add_connection(settings)


def edit_server(interface, settings):
    """Edit information for connecting to a server."""
    settings['common']['interface'] = interface
    settings['common']['name'] = 'WireGuard-Client-' + interface
    if not settings['wireguard']['private_key']:
        settings['wireguard']['private_key'] = _generate_private_key()

    connection = network.get_connection_by_interface_name(interface)
    network.edit_connection(connection, settings)
    network.reactivate_connection(connection.get_uuid())


def setup_server():
    """Setup a server connection that clients can connect to."""
    app = app_module.App.get('wireguard')
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    private_key = _generate_private_key()
    settings = {
        'common': {
            'name': 'WireGuard-Server-wg0',
            'type': setting_name,
            'zone': 'internal',
            'interface': 'wg0',
            'autoconnect': app.is_enabled(),
        },
        'ipv4': {
            'method': 'manual',
            'address': IP_TEMPLATE.format(1),
            'netmask': '255.255.255.0',
            'gateway': '',
            'dns': '',
            'second_dns': '',
        },
        'wireguard': {
            'private_key': private_key,
            'listen_port': 51820,
        }
    }
    network.add_connection(settings)
    logger.info('Created new WireGuard server connection')


def _get_next_available_ip_address(settings):
    """Get the next available IP address to allocate to a client."""
    allocated_ips = set()
    for peer_index in range(settings.get_peers_len()):
        peer = settings.get_peer(peer_index)
        for ip_index in range(peer.get_allowed_ips_len()):
            allowed_ip = peer.get_allowed_ip(ip_index)
            # We assume these are simple IP addresses but they can be subnets.
            allocated_ips.add(allowed_ip)

    for index in range(2, 254):
        ip_address = IP_TEMPLATE.format(index)
        if ip_address not in allocated_ips:
            return ip_address

    raise IndexError('Reached client limit')


def _server_connection():
    """Return a server connection. Create one if necessary."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    connection = network.get_connection_by_interface_name('wg0')
    if not connection:
        setup_server()

    for _ in range(10):
        # XXX: Improve this waiting by doing a synchronous D-Bus operation to
        # add network manager connection instead.
        time.sleep(1)
        connection = network.get_connection_by_interface_name('wg0')
        if connection:
            break

    if not connection:
        raise RuntimeError('Unable to create a server connection.')

    # Retrieve secrets so that when the connection is changed, secrets are
    # preserved properly.
    secrets = connection.get_secrets(setting_name)
    connection.update_secrets(setting_name, secrets)

    return connection


def add_client(public_key):
    """Add a permission for a client to connect our server."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    connection = _server_connection()
    settings = connection.get_setting_by_name(setting_name)
    peer, _ = settings.get_peer_by_public_key(public_key)
    if peer:
        raise ValueError('Peer with public key already exists')

    peer = nm.WireGuardPeer.new()
    peer.set_public_key(public_key, False)
    peer.set_persistent_keepalive(25)  # To keep NAT 'connections' alive
    peer.append_allowed_ip(_get_next_available_ip_address(settings), False)
    settings.append_peer(peer)
    connection.commit_changes(True)
    network.reactivate_connection(connection.get_uuid())


def remove_client(public_key):
    """Remove permission for a client to connect our server."""
    setting_name = nm.SETTING_WIREGUARD_SETTING_NAME
    connection = _server_connection()
    settings = connection.get_setting_by_name(setting_name)
    peer, peer_index = settings.get_peer_by_public_key(public_key)
    if not peer:
        raise KeyError('Client not found')

    settings.remove_peer(peer_index)
    connection.commit_changes(True)
    network.reactivate_connection(connection.get_uuid())


def build_client_config(client_ip: str, client_privkey: str,
                        client_pubkey: str, endpoint: str) -> str:
    """Generate WireGuard client config."""
    info = get_info()
    server_info = info['my_server']
    if not server_info:
        raise RuntimeError("WireGuard server not configured")

    server_pubkey = server_info['public_key']

    return f"""[Interface]
PrivateKey = {client_privkey}
Address = {client_ip}/32
DNS = 10.84.0.1

[Peer]
PublicKey = {server_pubkey}
AllowedIPs = 0.0.0.0/0
Endpoint = {endpoint}
PersistentKeepalive = 25"""



from unittest.mock import Mock, patch

import pytest

from plinth.modules.dynamicdns import gnudip

response_to_salt_request = """
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf8">
    <title>
        GnuDIP Update Server
    </title>
    <meta name="salt" content="gqEuCQYQWD">
    <meta name="time" content="1746978203">
    <meta name="sign" content="3d3b1c8ce32db470c6fd79a76f8dafb5">
</head>
<body></body>
</html>
"""

response_to_update_request = """
<html>
<head>
    <title>
        GnuDIP Update Server
    </title>
    <meta name="retc" content="0">
    <meta name="addr" content="24.81.172.128">
</head>
<body></body>
</html>
"""


def test_parse_meta_tags():
    """Test parsing meta tags from HTML content."""
    expected = {
        'salt': 'gqEuCQYQWD',
        'time': '1746978203',
        'sign': '3d3b1c8ce32db470c6fd79a76f8dafb5'
    }
    actual = gnudip._extract_content_from_meta_tags(response_to_salt_request)
    assert actual == expected


def test_check_required_keys_missing():
    """Test check_required_keys raises ValueError if key missing."""
    data = {'foo': 'bar'}
    with pytest.raises(ValueError) as excinfo:
        gnudip._check_required_keys(data, ['foo', 'baz'])

    assert "Missing required keys" in str(excinfo.value)


def test_check_required_keys_present():
    """Test check_required_keys with all keys present does not raise."""
    data = {'foo': 'bar', 'baz': 'qux'}
    gnudip._check_required_keys(data, ['foo', 'baz'])


def test_update_success():
    """Test GNU DIP update mechanism with HTTP protocol."""
    salt_resp = Mock()
    salt_resp.text = response_to_salt_request
    update_resp = Mock()
    update_resp.text = response_to_update_request

    with patch("plinth.modules.dynamicdns.gnudip.requests.get",
               side_effect=[salt_resp, update_resp]) as mock_get:
        result, addr = gnudip.update(server="http://www.2mbit.com:80",
                                     domain="gnudip.dyn.mpis.net",
                                     username="gnudip", password="password")
        assert result
        assert addr == "24.81.172.128"
        assert mock_get.call_count == 2



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django settings for test modules.
"""

import os

# Workaround for django-simple-captcha 0.5.6 not being compatible with
# Django 3.2. 0.5.14 is almost there in Debian. Workaround only until then.
import django.utils.encoding

django.utils.encoding.python_2_unicode_compatible = lambda x: x

TEST_DATA_DIR = os.path.dirname(os.path.abspath(__file__))

AXES_ENABLED = False

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache'
    }
}

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(TEST_DATA_DIR, 'plinth.sqlite3'),
    }
}

INSTALLED_APPS = [
    'axes',
    'captcha',
    'bootstrapform',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'oauth2_provider',
    'stronghold',
    'plinth',
]

# These are included here solely to suppress Django warnings
# during testing setup
MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

ROOT_URLCONF = 'plinth.tests.data.urls'

SECRET_KEY = 'django_tests_secret_key'

STRONGHOLD_PUBLIC_URLS = (r'^/o/', )

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'APP_DIRS': True,
    'DIRS': ['plinth/tests/data/templates/'],
}]

TIME_ZONE = 'UTC'

USE_TZ = True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the JSXC module."""

import copy

from django.http import Http404
from django.views.generic import TemplateView

import plinth.app as app_module
from plinth.middleware import CONTENT_SECURITY_POLICY
from plinth.modules.names.components import DomainName


class JsxcView(TemplateView):
    """A simple page to embed Javascript XMPP Client library."""

    template_name = 'jsxc_launch.html'
    headers: dict[str, str] = {}

    def __init__(self, **kwargs):
        """Initialize the view and set CSP."""
        super().__init__(**kwargs)
        csp = copy.copy(CONTENT_SECURITY_POLICY)
        csp['style-src'] = "'self' 'unsafe-inline'"
        self.headers['Content-Security-Policy'] = csp.get_header_value()

    def dispatch(self, request, *args, **kwargs):
        """Don't serve the view when app is disabled."""
        app = app_module.App.get('jsxc')
        if not app.is_enabled():
            raise Http404

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs) -> dict[str, object]:
        """Add domain information to view context."""
        context = super().get_context_data(*args, **kwargs)
        context['domain_name'] = DomainName.list_names()[0]
        return context

    def get(self, request, *args, **kwargs):
        """Handle GET request and return a response object."""
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context, headers=self.headers)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for users and groups.
"""

from django.utils.translation import gettext_lazy as _

tags = [_('Manage accounts'), _('App permissions')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Compare Debian package version numbers.
"""

from functools import total_ordering

from apt import apt_pkg


@total_ordering
class Version:
    """The version number of a Debian package."""

    def __init__(self, version: str):
        self.version = version

    def __eq__(self, other):
        return apt_pkg.version_compare(self.version, other.version) == 0

    def __lt__(self, other):
        return apt_pkg.version_compare(self.version, other.version) < 0



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure BIND server."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages, install
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('BIND enables you to publish your Domain Name System (DNS) information '
      'on the Internet, and to resolve DNS queries for your user devices on '
      'your network.'),
    format_lazy(
        _('Currently, on {box_name}, BIND is only used to resolve DNS queries '
          'for other machines on local network. It is also incompatible with '
          'sharing Internet connection from {box_name}.'),
        box_name=_(cfg.box_name)),
]


class BindApp(app_module.App):
    """FreedomBox app for Bind."""

    app_id = 'bind'

    _version = 4

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('BIND'), icon='fa-globe-w',
                               description=_description, manual_page='Bind',
                               tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-bind', info.name, info.icon, info.tags,
                              'bind:index',
                              parent_url_name='system:visibility', order=30)
        self.add(menu_item)

        packages = Packages('packages-bind', ['bind9'])
        self.add(packages)

        firewall = Firewall('firewall-bind', info.name, ports=['dns'],
                            is_external=False)
        self.add(firewall)

        daemon = Daemon(
            'daemon-bind', 'named', listen_ports=[(53, 'tcp6'), (53, 'udp6'),
                                                  (53, 'tcp4'), (53, 'udp4')])
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-bind',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup(old_version)
        if not old_version:
            self.enable()

    def force_upgrade(self, _packages):
        """Force upgrade the managed packages to resolve conffile prompt."""
        # Allow upgrades nay new version by keeping old configuration files
        install(['bind9'], force_configuration='old')
        return True



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import validate

clients = validate([{
    'name': _('Kiwix'),
    'platforms': [{
        'type': 'web',
        'url': '/kiwix'
    }]
}])

backup = {
    'data': {
        'directories': ['/var/lib/private/kiwix-server-freedombox/']
    },
    'services': ['kiwix-server-freedombox'],
    'delete_before_restore': ['/var/lib/private/kiwix-server-freedombox/']
}

tags = [
    _('Offline reader'),
    _('Archival'),
    _('Censorship resistance'),
    _('Wikipedia')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test the Let's Encrypt component for managing certificates.
"""

import contextlib
import random
from unittest.mock import call, patch

import pytest

from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.names.components import DomainName, DomainType


@pytest.fixture(name='empty_letsencrypt_list', autouse=True)
def fixture_empty_letsencrypt_list():
    """Remove all entries from Let's Encrypt component list."""
    LetsEncrypt._all = {}


@pytest.fixture(name='component')
def fixture_component():
    """Create a new component for testing."""
    reload_daemons = random.choice([True, False])
    component = LetsEncrypt(
        'test-component', domains=['valid.example', 'invalid.example'],
        daemons=['test-daemon'], should_copy_certificates=True,
        private_key_path='/etc/test-app/{domain}/private.path',
        certificate_path='/etc/test-app/{domain}/certificate.path',
        user_owner='test-user', group_owner='test-group',
        managing_app='test-app', reload_daemons=reload_daemons)
    assert component.reload_daemons == reload_daemons
    return component


@pytest.fixture(name='copy_certificate')
def fixture_copy_certificate():
    """Patch and return privileged.copy_certificate call."""
    with patch('plinth.modules.letsencrypt.privileged.copy_certificate'
               ) as copy_certificate:
        yield copy_certificate


@pytest.fixture(name='compare_certificate')
def fixture_compare_certificate():
    """Patch and return privileged.compare_certificate call."""
    with patch('plinth.modules.letsencrypt.privileged.compare_certificate'
               ) as compare_certificate:
        yield compare_certificate


@pytest.fixture(name='get_status')
def fixture_get_status():
    """Return patched letsencrypt.get_status() method."""
    domains = ['valid.example']
    with patch('plinth.modules.letsencrypt.get_status') as get_status:
        get_status.return_value = {
            'domains': {
                domain: {
                    'lineage': '/etc/letsencrypt/live/' + domain,
                    'validity': 'valid',
                }
                for domain in domains
            }
        }
        yield get_status


@pytest.fixture(name='domain_list')
def fixture_domain_list():
    """Return patch DomainName.list() method."""
    method = 'plinth.modules.names.components.DomainName.list'
    with patch(method) as domain_list:
        DomainType._all = {}
        DomainType('domain-type-1', 'type-1', 'url1', False)
        DomainType('domain-type-2', 'type-2', 'url1', True)
        domain1 = DomainName('domain-name-1', 'invalid1.example',
                             'domain-type-1', '__all__')
        domain2 = DomainName('domain-name-2', 'valid.example', 'domain-type-2',
                             '__all__')
        domain3 = DomainName('domain-name-3', 'invalid2.example',
                             'domain-type-2', '__all__')
        domain_list.return_value = [domain1, domain2, domain3]
        yield domain_list


def test_init_without_arguments():
    """Test that component is initialized with defaults properly."""
    component = LetsEncrypt('test-component')

    assert component.component_id == 'test-component'
    assert component.domains is None
    assert component.daemons is None
    assert not component.should_copy_certificates
    assert component.private_key_path is None
    assert component.certificate_path is None
    assert component.user_owner is None
    assert component.group_owner is None
    assert component.managing_app is None
    assert not component.reload_daemons
    assert len(component._all) == 1
    assert component._all['test-component'] == component


def test_init(component):
    """Test initializing the component."""
    assert component.domains == ['valid.example', 'invalid.example']
    assert component.daemons == ['test-daemon']
    assert component.should_copy_certificates
    assert component.private_key_path == '/etc/test-app/{domain}/private.path'
    assert component.certificate_path == \
        '/etc/test-app/{domain}/certificate.path'
    assert component.user_owner == 'test-user'
    assert component.group_owner == 'test-group'
    assert component.managing_app == 'test-app'


def test_init_values():
    """Test initializing with invalid values."""
    properties = {
        'private_key_path': 'test-private-key-path',
        'certificate_path': 'test-certificate-path',
        'user_owner': 'test-user',
        'group_owner': 'test-group',
        'managing_app': 'test-app'
    }
    LetsEncrypt('test-component', should_copy_certificates=True, **properties)
    for key in properties:
        new_properties = dict(properties)
        new_properties[key] = None
        with pytest.raises(ValueError):
            LetsEncrypt('test-component', should_copy_certificates=True,
                        **new_properties)


def test_domains():
    """Test getting domains."""
    component = LetsEncrypt('test-component', domains=lambda: ['test-domains'])
    assert component.domains == ['test-domains']


def test_list():
    """Test listing components."""
    component1 = LetsEncrypt('test-component1')
    component2 = LetsEncrypt('test-component2')
    assert set(LetsEncrypt.list()) == {component1, component2}


def _assert_copy_certificate_called(component, copy_certificate, domains):
    """Check that copy certificate calls have been made properly."""
    expected_calls = []
    for domain, domain_status in domains.items():
        if domain_status == 'valid':
            source_private_key_path = \
                '/etc/letsencrypt/live/{}/privkey.pem'.format(domain)
            source_certificate_path = \
                '/etc/letsencrypt/live/{}/fullchain.pem'.format(domain)
        else:
            source_private_key_path = '/etc/ssl/private/ssl-cert-snakeoil.key'
            source_certificate_path = '/etc/ssl/certs/ssl-cert-snakeoil.pem'

        private_key_path = '/etc/test-app/{}/private.path'.format(domain)
        certificate_path = '/etc/test-app/{}/certificate.path'.format(domain)
        expected_call = call(component.managing_app,
                             str(source_private_key_path),
                             str(source_certificate_path), private_key_path,
                             certificate_path, component.user_owner,
                             component.group_owner)
        expected_calls.append(expected_call)

    copy_certificate.assert_has_calls(expected_calls, any_order=True)


@contextlib.contextmanager
def _assert_restarted_daemons(component, daemons=None):
    """Check that a call has restarted the daemons of a component."""
    daemons = daemons if daemons is not None else component.daemons

    expected_calls = [call(daemon) for daemon in daemons]
    with patch('plinth.privileged.service.try_reload_or_restart'
               ) as try_reload_or_restart, patch(
                   'plinth.privileged.service.try_restart') as try_restart:
        yield

        if component.reload_daemons:
            try_reload_or_restart.assert_has_calls(expected_calls,
                                                   any_order=True)
            try_restart.assert_not_called()
        else:
            try_restart.assert_has_calls(expected_calls, any_order=True)
            try_reload_or_restart.assert_not_called()


def test_setup_certificates(copy_certificate, get_status, component):
    """Test that initial copying of certs for an app works."""
    with _assert_restarted_daemons(component):
        component.setup_certificates()

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
        'invalid.example': 'invalid'
    })


def test_setup_certificates_without_copy(copy_certificate, get_status,
                                         component):
    """Test that initial copying of certs for an app works."""
    component.should_copy_certificates = False
    with _assert_restarted_daemons(component):
        component.setup_certificates()

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_setup_certificates_with_app_domains(copy_certificate, get_status,
                                             component):
    """Test that initial copying of certs for an app works."""
    component._domains = ['irrelevant1.example', 'irrelevant2.example']
    with _assert_restarted_daemons(component):
        component.setup_certificates(
            app_domains=['valid.example', 'invalid.example'])

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
        'invalid.example': 'invalid'
    })


def test_setup_certificates_with_all_domains(domain_list, copy_certificate,
                                             get_status, component):
    """Test that initial copying for certs works when app domains is '*'."""
    component._domains = '*'
    with _assert_restarted_daemons(component):
        component.setup_certificates()

    _assert_copy_certificate_called(
        component, copy_certificate, {
            'valid.example': 'valid',
            'invalid1.example': 'invalid',
            'invalid2.example': 'invalid'
        })


def _assert_compare_certificate_called(component, compare_certificate,
                                       domains):
    """Check that compare certificate was called properly."""
    expected_calls = []
    for domain in domains:
        source_private_key_path = \
            '/etc/letsencrypt/live/{}/privkey.pem'.format(domain)
        source_certificate_path = \
            '/etc/letsencrypt/live/{}/fullchain.pem'.format(domain)
        private_key_path = '/etc/test-app/{}/private.path'.format(domain)
        certificate_path = '/etc/test-app/{}/certificate.path'.format(domain)
        expected_call = call(component.managing_app,
                             str(source_private_key_path),
                             str(source_certificate_path), private_key_path,
                             certificate_path)
        expected_calls.append(expected_call)

    compare_certificate.assert_has_calls(expected_calls, any_order=True)


def test_get_status(component, compare_certificate, get_status):
    """Test that getting domain status works."""
    compare_certificate.return_value = True
    assert component.get_status() == {
        'valid.example': 'valid',
        'invalid.example': 'self-signed'
    }
    _assert_compare_certificate_called(component, compare_certificate,
                                       ['valid.example'])


def test_get_status_outdate_copy(component, compare_certificate, get_status):
    """Test that getting domain status works with outdated copy."""
    compare_certificate.return_value = False
    assert component.get_status() == {
        'valid.example': 'outdated-copy',
        'invalid.example': 'self-signed'
    }
    _assert_compare_certificate_called(component, compare_certificate,
                                       ['valid.example'])


def test_get_status_without_copy(component, get_status):
    """Test that getting domain status works without copying."""
    component.should_copy_certificates = False
    assert component.get_status() == {
        'valid.example': 'valid',
        'invalid.example': 'self-signed'
    }


def test_on_certificate_obtained(copy_certificate, component):
    """Test that certificate obtained event handler works."""
    with _assert_restarted_daemons(component):
        component.on_certificate_obtained(
            ['valid.example', 'irrelevant.example'],
            '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })


def test_on_certificate_obtained_with_all_domains(copy_certificate, component):
    """Test that certificate obtained event handler works for app with
       all domains.
    """
    component._domains = '*'
    with _assert_restarted_daemons(component):
        component.on_certificate_obtained(
            ['valid.example'], '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })


def test_on_certificate_obtained_irrelevant(copy_certificate, component):
    """Test that certificate obtained event handler works with
       irrelevant domain.
    """
    with _assert_restarted_daemons(component, []):
        component.on_certificate_obtained(
            ['irrelevant.example'],
            '/etc/letsencrypt/live/irrelevant.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_obtained_without_copy(copy_certificate, component):
    """Test that certificate obtained event handler works without copying."""
    component.should_copy_certificates = False
    with _assert_restarted_daemons(component):
        component.on_certificate_obtained(
            ['valid.example'], '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_renewed(copy_certificate, component):
    """Test that certificate renewed event handler works."""
    with _assert_restarted_daemons(component):
        component.on_certificate_renewed(
            ['valid.example', 'irrelevant.example'],
            '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'valid',
    })


def test_on_certificate_renewed_irrelevant(copy_certificate, component):
    """Test that cert renewed event handler works for irrelevant domains."""
    with _assert_restarted_daemons(component, []):
        component.on_certificate_renewed(
            ['irrelevant.example'],
            '/etc/letsencrypt/live/irrelevant.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_renewed_without_copy(copy_certificate, component):
    """Test that certificate renewed event handler works without copying."""
    component.should_copy_certificates = False
    with _assert_restarted_daemons(component):
        component.on_certificate_renewed(
            ['valid.example'], '/etc/letsencrypt/live/valid.example/')
    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_revoked(copy_certificate, component):
    """Test that certificate revoked event handler works."""
    with _assert_restarted_daemons(component):
        component.on_certificate_revoked(
            ['valid.example', 'irrelevant.example'],
            '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'invalid',
    })


def test_on_certificate_revoked_irrelevant(copy_certificate, component):
    """Test that certificate revoked event handler works for
       irrelevant domains.
    """
    with _assert_restarted_daemons(component, []):
        component.on_certificate_revoked(
            ['irrelevant.example'],
            '/etc/letsencrypt/live/irrelevant.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_revoked_without_copy(copy_certificate, component):
    """Test that certificate revoked event handler works without copying."""
    component.should_copy_certificates = False
    with _assert_restarted_daemons(component):
        component.on_certificate_revoked(
            ['valid.example'], '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_deleted(copy_certificate, component):
    """Test that certificate deleted event handler works."""
    with _assert_restarted_daemons(component):
        component.on_certificate_deleted(
            ['valid.example', 'irrelevant.example'],
            '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {
        'valid.example': 'invalid',
    })


def test_on_certificate_deleted_irrelevant(copy_certificate, component):
    """Test that certificate deleted event handler works for
       irrelevant domains.
    """
    with _assert_restarted_daemons(component, []):
        component.on_certificate_deleted(
            ['irrelevant.example'],
            '/etc/letsencrypt/live/irrelevant.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})


def test_on_certificate_deleted_without_copy(copy_certificate, component):
    """Test that certificate deleted event handler works without copying."""
    component.should_copy_certificates = False
    with _assert_restarted_daemons(component):
        component.on_certificate_deleted(
            ['valid.example'], '/etc/letsencrypt/live/valid.example/')

    _assert_copy_certificate_called(component, copy_certificate, {})



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

from . import privileged

clients = [
    {
        'name': _('SOGo'),
        'platforms': [{
            'type': 'web',
            'url': '/SOGo/'
        }]
    },
    {
        'name':
            _('Thunderbird + SOGo connector'),
        'platforms': [{
            'type': 'download',
            'os': 'gnu-linux',
            'url': 'https://www.sogo.nu/download.html#/frontends'
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'https://www.sogo.nu/download.html#/frontends'
        }, {
            'type': 'download',
            'os': 'windows',
            'url': 'https://www.sogo.nu/download.html#/frontends'
        }]
    },
    {
        'name':
            _('DAVx5'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'at.bitfire.davdroid'),
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'at.bitfire.davdroid'),
        }]
    },
    {
        'name':
            _('GNOME Calendar'),
        'platforms': [{
            'type': 'package',
            'format': 'deb',
            'name': 'gnome-calendar'
        }]
    },
]

backup = {
    'data': {
        'files': [str(privileged.DB_BACKUP_FILE)],
    },
    'services': ['sogo'],
    'secrets': {
        'directories': [str(privileged.CONFIG_FILE)]
    },
}

tags = [
    _('Webmail'),
    _('Groupware'),
    _('Calendar'),
    _('Address book'),
    _('CalDAV'),
    _('CardDAV')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Roundcube module.
"""

from django.urls import re_path

from .views import RoundcubeAppView

urlpatterns = [
    re_path(r'^apps/roundcube/$', RoundcubeAppView.as_view(), name='index')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure a firewall."""

import contextlib
import logging

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg, menu
from plinth.daemon import Daemon
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules.backups.components import BackupRestore
from plinth.package import Packages, install
from plinth.utils import Version, format_lazy, import_from_gi

from . import manifest, privileged

gio = import_from_gi('Gio', '2.0')
glib = import_from_gi('GLib', '2.0')

_description = [
    format_lazy(
        _('Firewall is a security system that controls the incoming and '
          'outgoing network traffic on your {box_name}. Keeping a '
          'firewall enabled and properly configured reduces risk of '
          'security threat from the Internet.'), box_name=_(cfg.box_name))
]

_port_details: dict[str, list[str]] = {}

logger = logging.getLogger(__name__)

_DBUS_NAME = 'org.fedoraproject.FirewallD1'
_FIREWALLD_OBJECT = '/org/fedoraproject/FirewallD1'
_FIREWALLD_INTERFACE = 'org.fedoraproject.FirewallD1'
_ZONE_INTERFACE = 'org.fedoraproject.FirewallD1.zone'
_DIRECT_INTERFACE = 'org.fedoraproject.FirewallD1.direct'
_CONFIG_OBJECT = '/org/fedoraproject/FirewallD1/config'
_CONFIG_INTERFACE = 'org.fedoraproject.FirewallD1.config'
_CONFIG_SERVICE_INTERFACE = 'org.fedoraproject.FirewallD1.config.service'
_CONFIG_ZONE_INTERFACE = 'org.fedoraproject.FirewallD1.config.zone'
_CONFIG_DIRECT_INTERFACE = 'org.fedoraproject.FirewallD1.config.direct'


class FirewallApp(app_module.App):
    """FreedomBox app for Firewall."""

    app_id = 'firewall'

    _version = 5

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Firewall'),
                               icon='fa-shield', description=_description,
                               manual_page='Firewall', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-firewall', info.name, info.icon, info.tags,
                              'firewall:index',
                              parent_url_name='system:security', order=30)
        self.add(menu_item)

        packages = Packages('packages-firewall', ['firewalld', 'nftables'])
        self.add(packages)

        daemon = Daemon('daemon-firewall', 'firewalld')
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-firewall',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()

    def force_upgrade(self, packages):
        """Force upgrade firewalld to resolve conffile prompts."""
        if 'firewalld' not in packages:
            return False

        # Allow upgrade from 1.3.3 (bookworm) to 2.3.0 (trixie) and beyond 2.x.
        package = packages['firewalld']
        if Version(package['new_version']) > Version('3~'):
            return False

        install(['firewalld'], force_configuration='new')
        privileged.setup()
        return True

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()
        config = privileged.get_config()
        results.append(_diagnose_default_zone(config))
        results.append(_diagnose_firewall_backend(config))
        results.append(_diagnose_direct_passthroughs(config))
        return results


def _get_dbus_proxy(object, interface):
    """Return a DBusProxy for a given firewalld object and interface."""
    connection = gio.bus_get_sync(gio.BusType.SYSTEM)
    return gio.DBusProxy.new_sync(connection, gio.DBusProxyFlags.NONE, None,
                                  _DBUS_NAME, object, interface)


@contextlib.contextmanager
def ignore_dbus_error(dbus_error=None, service_error=None):
    try:
        yield
    except glib.Error as exception:
        parts = exception.message.split(':')
        if parts[0] != 'GDBus.Error':
            raise

        if (dbus_error and parts[1].strip()
                == 'org.freedesktop.DBus.Error.' + dbus_error):
            logger.error('Firewalld is not running.')
        elif (service_error and parts[2].strip() == service_error):
            logger.warning('Ignoring firewall exception: %s', service_error)
        else:
            raise


def parse_dbus_error(exception):
    """Parse a GDBus error."""
    parts = exception.message.split(':')
    if parts[0] != 'GDBus.Error' or \
       parts[1] != 'org.fedoraproject.FirewallD1.Exception':
        return None

    return parts[2].strip()


def reload():
    """Reload firewalld."""
    logger.info('Reloading firewalld')
    with ignore_dbus_error(dbus_error='ServiceUnknown'):
        proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _FIREWALLD_INTERFACE)
        proxy.reload()


def try_with_reload(operation):
    """Try an operation and retry after firewalld reload.

    When a service file is newly installed into /usr/lib/firewalld/services,
    it's information can be immediately queried but the service can't be
    added/removed from a zone. A firewalld reload is necessary. So, try an
    operation and if it fails with INVALID_SERVICE error, reload firewalld and
    try again.

    """
    try:
        operation()
    except glib.Error as exception:
        error = parse_dbus_error(exception)
        if error != 'INVALID_SERVICE':
            raise

        reload()
        operation()


def get_enabled_services(zone):
    """Return the status of various services currently enabled."""
    with ignore_dbus_error(dbus_error='ServiceUnknown'):
        zone_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _ZONE_INTERFACE)
        return zone_proxy.getServices('(s)', zone)

    return []  # When firewalld is not running


def get_port_details(service_port):
    """Return the port types and numbers for a service port."""
    try:
        return _port_details[service_port]
    except KeyError:
        config = _get_dbus_proxy(_CONFIG_OBJECT, _CONFIG_INTERFACE)
        try:
            service_path = config.getServiceByName('(s)', service_port)
        except glib.Error:
            return []  # Don't cache the error result

        service = _get_dbus_proxy(service_path, _CONFIG_SERVICE_INTERFACE)
        _port_details[service_port] = service.getPorts()
        return _port_details[service_port]


def get_interfaces(zone):
    """Return the list of interfaces in a zone."""
    with ignore_dbus_error(dbus_error='ServiceUnknown'):
        zone_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _ZONE_INTERFACE)
        return zone_proxy.getInterfaces('(s)', zone)

    return []  # When firewalld is not running


def add_service(port, zone):
    """Enable a service in firewall."""
    with ignore_dbus_error(dbus_error='ServiceUnknown'):
        zone_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _ZONE_INTERFACE)
        with ignore_dbus_error(service_error='ALREADY_ENABLED'):
            zone_proxy.addService('(ssi)', zone, port, 0)

        config = _get_dbus_proxy(_CONFIG_OBJECT, _CONFIG_INTERFACE)
        zone_path = config.getZoneByName('(s)', zone)
        config_zone = _get_dbus_proxy(zone_path, _CONFIG_ZONE_INTERFACE)
        with ignore_dbus_error(service_error='ALREADY_ENABLED'):
            config_zone.addService('(s)', port)


def remove_service(port, zone):
    """Remove a service in firewall."""
    with ignore_dbus_error(dbus_error='ServiceUnknown'):
        zone_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _ZONE_INTERFACE)
        with ignore_dbus_error(service_error='NOT_ENABLED'):
            zone_proxy.removeService('(ss)', zone, port)

        config = _get_dbus_proxy(_CONFIG_OBJECT, _CONFIG_INTERFACE)
        zone_path = config.getZoneByName('(s)', zone)
        config_zone = _get_dbus_proxy(zone_path, _CONFIG_ZONE_INTERFACE)
        with ignore_dbus_error(service_error='NOT_ENABLED'):
            config_zone.removeService('(s)', port)


def add_passthrough(ipv, *args):
    """Add a direct rule with passthrough to ip(6)tables."""
    direct_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _DIRECT_INTERFACE)
    if not direct_proxy.queryPassthrough('(sas)', ipv, args):
        direct_proxy.addPassthrough('(sas)', ipv, args)

    config_direct = _get_dbus_proxy(_CONFIG_OBJECT, _CONFIG_DIRECT_INTERFACE)
    if not config_direct.queryPassthrough('(sas)', ipv, args):
        config_direct.addPassthrough('(sas)', ipv, args)


def remove_passthrough(ipv, *args):
    """Add a direct rule with passthrough to ip(6)tables."""
    direct_proxy = _get_dbus_proxy(_FIREWALLD_OBJECT, _DIRECT_INTERFACE)
    if direct_proxy.queryPassthrough('(sas)', ipv, args):
        direct_proxy.removePassthrough('(sas)', ipv, args)

    config_direct = _get_dbus_proxy(_CONFIG_OBJECT, _CONFIG_DIRECT_INTERFACE)
    if config_direct.queryPassthrough('(sas)', ipv, args):
        config_direct.removePassthrough('(sas)', ipv, args)


def _diagnose_default_zone(
        config: privileged.FirewallConfig) -> DiagnosticCheck:
    """Diagnose whether the default zone is external."""
    check_id = 'firewall-default-zone'
    description = gettext_noop('Default zone is external')
    result = Result.PASSED if config[
        'default_zone'] == 'external' else Result.FAILED
    return DiagnosticCheck(check_id, description, result)


def _diagnose_firewall_backend(
        config: privileged.FirewallConfig) -> DiagnosticCheck:
    """Diagnose whether the firewall backend is nftables."""
    check_id = 'firewall-backend'
    description = gettext_noop('Firewall backend is nftables')
    result = Result.PASSED if config['backend'] == 'nftables' \
        else Result.FAILED
    return DiagnosticCheck(check_id, description, result)


def _diagnose_direct_passthroughs(
        config: privileged.FirewallConfig) -> DiagnosticCheck:
    """Diagnose direct passthroughs for local service protection.

    Currently, we just check that the number of passthroughs is at least 12,
    which are the number that are added by firewall's setup.
    """
    check_id = 'firewall-direct-passthroughs'
    description = gettext_noop('Direct passthrough rules exist')
    result = Result.PASSED if len(
        config['passthroughs']) >= 12 else Result.FAILED
    return DiagnosticCheck(check_id, description, result)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for TiddlyWiki.

This is a FreedomBox-native implementation of a TiddlyWiki Nest.
This app doesn't install any Debian packages.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('TiddlyWiki is an interactive application that runs entirely in the '
          'web browser. Each wiki is a self-contained HTML file stored on your'
          ' {box_name}. Instead of writing long wiki pages, TiddlyWiki '
          'encourages you to write several short notes called Tiddlers and '
          'link them together into a dense graph.'), box_name=_(cfg.box_name)),
    _('It is a versatile application with a wide variety of use cases - '
      'non-linear notebook, website, personal knowledge base, task and project'
      ' management system, personal diary etc. Plugins can extend the '
      'functionality of TiddlyWiki. Encrypting individual tiddlers or '
      'password-protecting a wiki file is possible from within the '
      'application.'),
    format_lazy(
        _('TiddlyWiki is downloaded from {box_name} website and not from '
          'Debian. Wikis need to be upgraded to newer version manually.'),
        box_name=_(cfg.box_name)),
    format_lazy(
        _('Wikis are not public by default, but they can be downloaded for '
          'sharing or publishing. They can be edited by <a href="{users_url}">'
          'any user</a> on {box_name} belonging to the wiki group. '
          'Simultaneous editing is not supported.'), box_name=_(cfg.box_name),
        users_url=reverse_lazy('users:index')),
    _('Create a new wiki or upload your existing wiki file to get started.')
]


class TiddlyWikiApp(app_module.App):
    """FreedomBox app for TiddlyWiki."""

    app_id = 'tiddlywiki'

    _version = 1

    def __init__(self):
        """Create components for the app."""
        super().__init__()

        groups = {'wiki': _('View and edit wiki applications')}

        info = app_module.Info(self.app_id, self._version,
                               name=_('TiddlyWiki'),
                               icon_filename='tiddlywiki',
                               description=_description,
                               manual_page='TiddlyWiki',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-tiddlywiki', info.name, info.icon_filename,
                              info.tags, 'tiddlywiki:index',
                              parent_url_name='apps')
        self.add(menu_item)

        # The shortcut is a simple directory listing provided by Apache server.
        # Expecting a large number of wiki files, so creating a shortcut for
        # each file (like in ikiwiki's case) will crowd the front page.
        shortcut = frontpage.Shortcut(
            'shortcut-tiddlywiki', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            url='/tiddlywiki/', clients=info.clients, tags=info.tags,
            login_required=True, allowed_groups=list(groups))
        self.add(shortcut)

        dropin_configs = DropinConfigs('dropin-configs-tiddlywiki', [
            '/etc/apache2/conf-available/tiddlywiki-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-tiddlywiki', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-tiddlywiki', 'tiddlywiki-freedombox')
        self.add(webserver)

        users_and_groups = UsersAndGroups('users-and-groups-tiddlywiki',
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-tiddlywiki',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version=None):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """Purge directory with all the wikis."""
        super().uninstall()
        privileged.uninstall()


def get_wiki_list():
    """List all the TiddlyWiki files."""
    return sorted([
        path.name for path in privileged.wiki_dir.iterdir()
        if path.suffix == '.html'
    ])



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('Zoph'),
    'platforms': [{
        'type': 'web',
        'url': '/zoph/',
    }]
}]

backup = {
    'data': {
        'files': ['/var/lib/plinth/backups-data/zoph-database.sql'],
        'directories': ['/var/lib/zoph/']
    },
    'secrets': {
        'files': [
            '/etc/zoph.ini',
            '/etc/dbconfig-common/zoph.conf',
        ],
    }
}

tags = [_('Photo'), _('Organizer'), _('Web sharing')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Utilities for performing application setup operations.
"""

import itertools
import logging
import threading
import time
from collections import defaultdict
from typing import Union

import apt
from django.utils.translation import gettext_noop

import plinth
from plinth import app as app_module
from plinth.diagnostic_check import Result
from plinth.errors import MissingPackageError
from plinth.package import Packages
from plinth.signals import post_setup

from . import operation as operation_module
from . import package
from .privileged import packages as packages_privileged

logger = logging.getLogger(__name__)

is_first_setup_running = False
_is_shutting_down = False

thread_local_storage = threading.local()


def run_setup_on_app(app_id, allow_install=True, rerun=False):
    """Execute the setup process in a thread."""
    # App is already up-to-date
    app = app_module.App.get(app_id)
    current_version = app.get_setup_version()
    if not rerun and current_version >= app.info.version:
        return

    if not current_version:
        name = gettext_noop('Installing app')
    else:
        name = gettext_noop('Updating app')

    logger.debug('Creating operation to setup app: %s', app_id)
    show_notification = show_message = (current_version
                                        or not app.info.is_essential)
    return operation_module.manager.new(
        f'{app_id}-setup', app_id, name, _run_setup_on_app,
        [app, current_version], show_message=show_message,
        show_notification=show_notification,
        thread_data={'allow_install': allow_install})


def _run_setup_on_app(app, current_version, repair: bool = False):
    """Execute the setup process."""
    logger.info('Setup run: %s', app.app_id)
    exception_to_update: Exception | None = None
    message = None
    try:
        current_version = app.get_setup_version()

        if current_version != 0:
            # Check if this app needs force_upgrade. If it is needed, but not
            # yet supported for the new version of the package, then an
            # exception will be raised, so that we do not run setup.
            package.refresh_package_lists()
            force_upgrader = ForceUpgrader.get_instance()
            force_upgrader.attempt_upgrade_for_app(app.app_id)

        app.setup(old_version=current_version)
        app.set_setup_version(app.info.version)
        post_setup.send_robust(sender=app.__class__, module_name=app.app_id)
    except Exception as exception:
        exception_to_update = exception
        if not current_version:
            message = gettext_noop('Error installing app: {exception}')
        elif repair:
            message = gettext_noop('Error repairing app: {exception}')
        else:
            message = gettext_noop('Error updating app: {exception}')
    else:
        if not current_version:
            message = gettext_noop('App installed.')
        elif repair:
            return
        else:
            message = gettext_noop('App updated')

    logger.info('Setup completed: %s: %s %s', app.app_id, message,
                exception_to_update)
    operation = operation_module.Operation.get_operation()
    operation.on_update(message, exception_to_update)


def run_repair_on_app(app_id, create_operation=True):
    """Execute the repair process in a thread.

    In case this is called from within another operation, creating a new
    operation can be skipped.
    """
    app = app_module.App.get(app_id)
    current_version = app.get_setup_version()
    if not current_version:
        logger.warning('App %s is not installed, cannot repair', app_id)
        return

    if create_operation:
        logger.debug('Creating operation to repair app: %s', app_id)
        return operation_module.manager.new(f'{app_id}-repair', app_id,
                                            gettext_noop('Repairing app'),
                                            _run_repair_on_app, [app],
                                            show_message=True,
                                            show_notification=True)

    # Re-use existing operation.
    try:
        operation = operation_module.Operation.get_operation()
    except AttributeError:
        raise RuntimeError(
            'run_repair_on_app: Expected an existing operation.')

    # XXX: Ugly hack to re-use operation from another app.
    operation.app_id = app_id
    _run_repair_on_app(app)


def _run_repair_on_app(app: app_module.App):
    """Execute the repair process."""
    logger.info('Repair run: %s', app.app_id)
    message = None
    operation = operation_module.Operation.get_operation()

    # Always re-run diagnostics first for this app, to ensure results are
    # current.
    checks = []
    try:
        checks = app.diagnose()
    except Exception as exception:
        logger.error('Error running %s diagnostics - %s', app.app_id,
                     exception)
        message = gettext_noop('Error running diagnostics: {error}').format(
            error=exception)
        operation.on_update(message, exception)
        return

    # Filter for checks that have failed.
    failed_checks = []
    for check in checks:
        if check.result in [Result.FAILED, Result.WARNING]:
            failed_checks.append(check)

    if not failed_checks:
        logger.warning('Skipping repair for %s: no failed checks', app.app_id)
        message = gettext_noop('Skipping repair, no failed checks')
        operation.on_update(message, None)
        return

    try:
        should_rerun_setup = app.repair(failed_checks)
    except Exception as exception:
        logger.error('Repair error: %s: %s %s', app.app_id, message, exception)
        message = gettext_noop('Error repairing app: {error}').format(
            error=exception)
        operation.on_update(message, exception)
        return

    if should_rerun_setup:
        message = gettext_noop('Re-running setup to complete repairs')
        operation.on_update(message, None)
        current_version = app.get_setup_version()
        _run_setup_on_app(app, current_version, True)

    logger.info('Repair completed: %s', app.app_id)

    # Check for errors in thread local storage
    message = gettext_noop('App repaired.')
    errors = retrieve_error_messages()
    exceptions = None
    if errors:
        message = gettext_noop('App repair completed with errors:\n')
        error_message = ''
        for error in errors:
            message += str(error) + '\n'
            error_message += str(error) + '\n'

        exceptions = Exception(error_message)

    operation.on_update(message, exceptions)


def run_uninstall_on_app(app_id):
    """Execute the uninstall process in a thread."""
    # App is already uninstalled
    app = app_module.App.get(app_id)
    if not app.get_setup_version():
        return

    logger.debug('Creating operation to uninstall app: %s', app_id)
    return operation_module.manager.new(f'{app_id}-uninstall', app_id,
                                        gettext_noop('Uninstalling app'),
                                        _run_uninstall_on_app, [app],
                                        show_notification=True)


def _run_uninstall_on_app(app):
    """Execute the uninstall process."""
    logger.info('Uninstall run: %s', app.app_id)
    exception_to_update = None
    message = None
    try:
        app.disable()
        app.uninstall()
        app.set_setup_version(0)
    except Exception as exception:
        exception_to_update = exception
        message = gettext_noop('Error uninstalling app: {error}').format(
            error=exception)
    else:
        message = gettext_noop('App uninstalled.')

    logger.info('Uninstall completed: %s: %s %s', app.app_id, message,
                exception_to_update)
    operation = operation_module.Operation.get_operation()
    operation.on_update(message, exception_to_update)


def stop():
    """Set a flag to indicate that the setup process must stop."""
    global _is_shutting_down
    _is_shutting_down = True

    force_upgrader = ForceUpgrader.get_instance()
    force_upgrader.shutdown()

    dpkg_handler = DpkgHandler.get_instance()
    dpkg_handler.shutdown()


def setup_apps(app_ids=None, essential=False, allow_install=True):
    """Run setup on selected or essential apps."""
    logger.info(
        'Running setup for apps, essential - %s, '
        'selected apps - %s', essential, app_ids)
    for app in app_module.App.list():
        if essential and not app.info.is_essential:
            continue

        if app_ids and app.app_id not in app_ids:
            continue

        operation = run_setup_on_app(app.app_id, allow_install=allow_install)
        if operation:
            operation.join()


def list_dependencies(app_ids=None, essential=False):
    """Print list of packages required by selected or essential apps."""
    for app in app_module.App.list():
        if essential and not app.info.is_essential:
            continue

        if app_ids and app.app_id not in app_ids and \
           '*' not in app_ids:
            continue

        for component in app.get_components_of_type(Packages):
            for package_expression in component.package_expressions:
                print(package_expression)


def run_setup_on_startup():
    """Run setup with retry till it succeeds."""
    is_first_setup = any((app for app in app_module.App.list()
                          if app.info.is_essential and app.needs_setup()))

    sleep_time = 10
    while True:
        try:
            if is_first_setup:
                logger.info('Running first setup.')
                _run_first_setup()
                break
            else:
                logger.info('Running regular setup.')
                _run_regular_setup()
                break
        except Exception as ex:
            logger.warning('Unable to complete setup: %s', ex)
            logger.info('Will try again in {} seconds'.format(sleep_time))
            time.sleep(sleep_time)
            if _is_shutting_down:
                break

    logger.info('Setup finished.')


def _run_first_setup():
    """Run setup on essential apps on first setup."""
    global is_first_setup_running
    is_first_setup_running = True
    run_setup_on_apps(None, allow_install=False)
    is_first_setup_running = False


def _run_regular_setup():
    """Run setup on all apps also installing required packages."""
    app_ids = _get_apps_for_regular_setup()
    run_setup_on_apps(app_ids, allow_install=True)


def _get_apps_for_regular_setup():

    def is_setup_required(app):
        """Setup is required for:
        1. essential apps that are not up-to-date
        2. non-essential app that are installed and need updates
        """
        if (app.info.is_essential and app.get_setup_state()
                != app_module.App.SetupState.UP_TO_DATE):
            return True

        if app.get_setup_state() == app_module.App.SetupState.NEEDS_UPDATE:
            return True

        return False

    return [
        app.app_id for app in app_module.App.list() if is_setup_required(app)
    ]


def run_setup_on_apps(app_ids, allow_install=True):
    """Run setup on the given list of apps.

    apps is the list of apps to run setup on. If None is given, run setup on
    all essential apps only.

    allow_install with or without package installation. When setting up
    essential apps, installing packages is not required as FreedomBox
    (Plinth) itself has dependencies on all essential apps.

    """
    try:
        if not app_ids:
            setup_apps(essential=True, allow_install=allow_install)
        else:
            setup_apps(app_ids, allow_install=allow_install)
    except Exception as exception:
        logger.exception('Error running setup - %s', exception)
        raise


class ForceUpgrader():
    """Find and upgrade packages by force when conffile prompt is needed.

    Primary duty here is take responsibility of packages that will be rejected
    by unattended-upgrades for upgrades. Only packages that apps manage are
    handled as that is only case in which we know how to deal with
    configuration file changes. Roughly, we will read old configuration file,
    force upgrade the package with new configuration and then apply the old
    preferences on it.

    When upgrades packages, many things have to be kept in mind. The following
    is the understanding on how unattended-upgrades does its upgrades. These
    precautions need to be taken into account whenever relevant.

    Checks done by unattended upgrades before upgrading:
    - Check if today is a configured patch day.
    - Check if running on development release.
    - Cache does not have broken packages.
    - Whether system (and hence this process itself) is shutting down.
    - System is on AC Power (if configured).
    - Whether system is on a metered connection.
    - Package system is locked.
    - Check if dpkg journal is dirty and fix it.
    - Take care that upgrade process does not terminate/crash self.
    - Check if cache has broken dependencies.
    - Remove auto removable packages.
    - Download all packages before upgrade
    - Perform conffile prompt checks
      - Ignore if dpkg options has --force-confold or --force-confnew

    Unattended upgrades checks for configuration file prompts as follows:
    - Package downloaded successfully.
    - Package is complete.
    - File is available.
    - Package is trusted.
    - Package is not in denylist.
    - Package extension is .deb.
    - Get conffiles values from control data of .deb.
    - Read from /var/lib/dpkg/status, Package and Conffiles section.
    - Read each conffile from system and compute md5sum.
    - If new package md5sum == dpkg status md5 == file's md5 then no conffile
      prompt.
    - If conffiles in new package not found in dpkg status but found on system
      with different md5sum, conffile prompt is available.

    Filtering of packages to upgrade is done as follows:
    - Packages is upgradable.
    - Remove packages from denylist
    - Remove packages unless in allowlist or allowlist is empty
    - Package in allowed origins
    - Don't consider packages that require restart (if configured).
    - Package dependencies satisfy origins, denylist and allowlist.
    - Order packages alphabetically (order changes when upgrading in minimal
      steps).

    The upgrade process looks like:
    - Check for install while shutting down.
    - Check for dry run.
    - Mail list of changes.
    - Upgrade in minimal steps.
    - Clean downloaded packages.

    """

    _instance = None
    _run_lock = threading.Lock()
    _wait_event = threading.Event()

    UPGRADE_ATTEMPTS = 12

    UPGRADE_ATTEMPT_WAIT_SECONDS = 30 * 60

    class TemporaryFailure(Exception):
        """Raised when upgrade fails but can be tried again immediately."""

    class PermanentFailure(Exception):
        """Raised when upgrade fails and there is nothing more we wish to do.
        """

    @classmethod
    def get_instance(cls):
        """Return a single instance of a the class."""
        if not cls._instance:
            cls._instance = ForceUpgrader()

        return cls._instance

    def __init__(self):
        """Initialize the force upgrader."""
        if plinth.cfg.develop:
            self.UPGRADE_ATTEMPT_WAIT_SECONDS = 10

    def on_package_cache_updated(self):
        """Trigger upgrades when notified about changes to package cache.

        Call the upgrade process guaranteeing that it will not run more than
        once simultaneously.

        """
        if not self._run_lock.acquire(blocking=False):
            logger.info('Package upgrade already in process')
            return

        try:
            self._run_upgrade()
        finally:
            self._run_lock.release()

    def _run_upgrade(self):
        """Attempt the upgrade process multiple times.

        This method is guaranteed to not to run more than once simultaneously.

        """
        for _ in range(self.UPGRADE_ATTEMPTS):
            try:
                logger.info('Attempting to perform upgrade')
                self._attempt_upgrade()
                logger.info('Completed upgrade')
                return
            except self.TemporaryFailure as exception:
                logger.info('Cannot perform upgrade now: %s', exception)
            except self.PermanentFailure as exception:
                logger.error('Upgrade failed: %s', exception)
                return
            except Exception as exception:
                # Assume all other errors are temporary
                logger.exception('Unknown exception: %s', exception)

            logger.info('Waiting for %s seconds before attempting upgrade',
                        self.UPGRADE_ATTEMPT_WAIT_SECONDS)
            if self._wait_event.wait(self.UPGRADE_ATTEMPT_WAIT_SECONDS):
                logger.info('Stopping upgrade attempts due to shutdown')
                return

        logger.info('Giving up on upgrade after too many retries')

    def shutdown(self):
        """If we are sleeping for next attempt, cancel it.

        If we are actually upgrading packages, do nothing.
        """
        self._wait_event.set()

    def _attempt_upgrade(self):
        """Attempt to perform an upgrade.

        Raise TemporaryFailure if upgrade can't be performed now.

        Raise PermanentFailure if upgrade can't be performed until something
        with the system state changes. We don't want to try again until
        notified of further package cache changes.

        Return nothing otherwise.

        """
        if _is_shutting_down:
            raise self.PermanentFailure('Service is shutting down')

        if packages_privileged.is_package_manager_busy():
            raise self.TemporaryFailure('Package manager is busy')

        apps = self._get_list_of_apps_to_force_upgrade()
        logger.info(
            'Apps needing conffile upgrades: %s', ', '.join(
                [str(app_module.App.get(app_id).info.name) for app_id in apps])
            or 'None')

        need_retry = False
        for app_id, packages in apps.items():
            app = app_module.App.get(app_id)
            try:
                logger.info('Force upgrading app: %s', app.info.name)
                if self._run_force_upgrade_as_operation(app, packages):
                    logger.info('Successfully force upgraded app: %s',
                                app.info.name)
                else:
                    logger.info('Ignored force upgrade for app: %s',
                                app.info.name)
            except Exception as exception:
                logger.exception('Error running force upgrade: %s', exception)
                need_retry = True
                # Continue trying to upgrade other apps

        if need_retry:
            raise self.TemporaryFailure('Some apps failed to force upgrade.')

    def attempt_upgrade_for_app(self, app_id):
        """Attempt to perform an upgrade for specified app.

        Raise TemporaryFailure if upgrade can't be performed now.

        Raise PermanentFailure if upgrade can't be performed until something
        with the system state changes. We don't want to try again until
        notified of further package cache changes.

        Return True if upgrade was performed successfully.

        Return False if upgrade is not needed.

        """
        if _is_shutting_down:
            raise self.PermanentFailure('Service is shutting down')

        if packages_privileged.is_package_manager_busy():
            raise self.TemporaryFailure('Package manager is busy')

        apps = self._get_list_of_apps_to_force_upgrade()
        if app_id not in apps:
            logger.info('App %s does not need force upgrade', app_id)
            return False

        packages = apps[app_id]
        app = app_module.App.get(app_id)
        try:
            logger.info('Force upgrading app: %s', app.info.name)
            if app.force_upgrade(packages):
                logger.info('Successfully force upgraded app: %s',
                            app.info.name)
                return True
            else:
                logger.warning('Ignored force upgrade for app: %s',
                               app.info.name)
                raise self.TemporaryFailure(
                    'Force upgrade is needed, but not yet implemented for new '
                    f'version of app: {app_id}')
        except Exception as exception:
            logger.exception('Error running force upgrade: %s', exception)
            raise self.TemporaryFailure(
                f'App {app_id} failed to force upgrade.')

    def _run_force_upgrade_as_operation(self, app, packages):
        """Start an operation for force upgrading."""
        name = gettext_noop('Updating app packages')
        operation = operation_module.manager.new(f'{app.app_id}-force-upgrade',
                                                 app.app_id, name,
                                                 app.force_upgrade, [packages],
                                                 show_message=False,
                                                 show_notification=False)
        return operation.join()  # Wait for completion, raise Exception

    def _get_list_of_apps_to_force_upgrade(self):
        """Return a list of app on which to run force upgrade."""
        packages = self._get_list_of_upgradable_packages()
        if not packages:  # No packages to upgrade
            return {}

        package_names = [package.name for package in packages]
        logger.info('Packages available for upgrade: %s',
                    ', '.join(package_names))

        managed_packages, package_apps_map = self._filter_managed_packages(
            package_names)
        if not managed_packages:
            return {}

        conffile_packages = package.filter_conffile_prompt_packages(
            managed_packages)
        logger.info('Packages that need conffile upgrades: %s',
                    conffile_packages)

        apps = defaultdict(dict)
        for package_name in conffile_packages:
            for app_id in package_apps_map[package_name]:
                apps[app_id][package_name] = conffile_packages[package_name]

        return apps

    @staticmethod
    def _get_list_of_upgradable_packages():
        """Return list of packages that can be upgraded."""
        cache = apt.cache.Cache()
        return [package for package in cache if package.is_upgradable]

    @staticmethod
    def _filter_managed_packages(packages):
        """Filter out packages that apps don't force upgrade.

        Return packages in the list that are managed by at least one app that
        can perform force upgrade.

        """
        package_apps_map = defaultdict(set)
        upgradable_packages = set()
        for app in app_module.App.list():
            if not getattr(app, 'force_upgrade', None):
                # App does not implement force upgrade
                continue

            if app.get_setup_state() == app_module.App.SetupState.NEEDS_SETUP:
                # If an app is not installed don't considered it. If an app
                # needs an update, it may have to do a force upgrade before
                # running app version update. This is because the app version
                # update process will include installing packages that will
                # fail due to pending configuration file updates.
                continue

            for component in app.get_components_of_type(Packages):
                upgradable_packages.update(component.possible_packages)

                for managed_package in component.possible_packages:
                    package_apps_map[managed_package].add(app.app_id)

        return upgradable_packages.intersection(
            set(packages)), package_apps_map


def on_package_cache_updated():
    """Called by D-Bus service when apt package cache is updated."""
    force_upgrader = ForceUpgrader.get_instance()
    force_upgrader.on_package_cache_updated()


class DpkgHandler:
    """Find and rerun setup for apps after a dpkg operation is completed.

    This is needed in a couple of situations:

    1) Some Debian packages don't manage the database used by the package. When
    these packages are updated, their database schema is left at an older
    version and service might become unavailable. FreedomBox can perform the
    database schema upgrade. However, FreedomBox needs to know when a package
    has been updated so that database schema can be upgraded.

    2) A package is installed but FreedomBox has not modified its
    configuration. Newer version of package becomes available with a new
    configuration file. Since the original configuration file has not changed
    at all, the new configuration file overwrites the old one and
    unattended-upgrades deals with this case. Now, say, the configuration file
    modifies some defaults that FreedomBox expects things might break. In this
    case, FreedomBox can apply the require configuration changes but it needs
    to notified as soon as the package has been updated.

    When apt runs dpkg, after the operation is completed it triggers commands
    listed under the configuration 'Dpkg::Post-Invoke'. This in turn calls this
    class via a DBus notification. Here, we iterate through all the apps. If an
    app is currently installed and interested in rerunning setup after dpkg
    operations, then its setup is rerun. Interest is expressed using the
    'rerun_setup_on_upgrade' flag on the Package() component. If all packages
    of the app have not be upgraded since the last check, we skip the
    operation.
    """

    _instance: Union['DpkgHandler', None] = None
    _run_lock = threading.Lock()
    _wait_event = threading.Event()

    HANDLE_ATTEMPTS: int = 12

    HANDLE_ATTEMPT_WAIT_SECONDS: int = 30 * 60

    class TemporaryFailure(Exception):
        """Raised when post-dpkg operations fails but can be tried again."""

    class PermanentFailure(Exception):
        """Raised when post-dpkg operations fails.

        And there is nothing more we wish to do.
        """

    @classmethod
    def get_instance(cls) -> 'DpkgHandler':
        """Return a single instance of a the class."""
        if not cls._instance:
            cls._instance = DpkgHandler()

        return cls._instance

    def __init__(self) -> None:
        """Initialize the dpkg handler."""
        if plinth.cfg.develop:
            self.HANDLE_ATTEMPT_WAIT_SECONDS = 10

    def on_dpkg_invoked(self) -> None:
        """Trigger post-dpkg operations when notified about dpkg invocation.

        Call the post-dpkg operations guaranteeing that it will not run more
        than once simultaneously.
        """
        if not self._run_lock.acquire(blocking=False):
            logger.info('Post dpkg operations already in process')
            return

        try:
            self._on_dpkg_invoked()
        finally:
            self._run_lock.release()

    def _on_dpkg_invoked(self) -> None:
        """Attempt the post-dpkg operations multiple times.

        This method is guaranteed to not to run more than once simultaneously.
        """
        for _ in range(self.HANDLE_ATTEMPTS):
            try:
                logger.info('Attempting to perform post-dpkg operations')
                self._attempt_post_invoke()
                logger.info('Completed post-dpkg operations')
                return
            except self.TemporaryFailure as exception:
                logger.info('Cannot perform post-dpkg operations now: %s',
                            exception)
            except self.PermanentFailure as exception:
                logger.error('Post-dpkg operations failed: %s', exception)
                return
            except Exception as exception:
                # Assume all other errors are temporary
                logger.exception('Unknown exception: %s', exception)

            logger.info(
                'Waiting for %s seconds before attempting post-dpkg '
                'operations', self.HANDLE_ATTEMPT_WAIT_SECONDS)
            if self._wait_event.wait(self.HANDLE_ATTEMPT_WAIT_SECONDS):
                logger.info(
                    'Stopping post-dpkg operation attempts due to shutdown')
                return

        logger.info('Giving up on post-dpkg operations after too many retries')

    def _attempt_post_invoke(self) -> None:
        """Run post-dpkg invoke operations on all interested app."""
        if _is_shutting_down:
            raise self.PermanentFailure('Service is shutting down')

        if packages_privileged.is_package_manager_busy():
            raise self.TemporaryFailure('Package manager is busy')

        for app in app_module.App.list():
            self._post_invoke_on_app(app)

    def _post_invoke_on_app(self, app: app_module.App) -> None:
        """Run post-dpkg invoke operations on a single app."""
        components = list(app.get_components_of_type(Packages))
        app_interested = any(
            (component.rerun_setup_on_upgrade for component in components))
        if not app_interested:
            # App is not interested in re-running setup after a package has
            # been updated.
            return

        if app.get_setup_state() == app_module.App.SetupState.NEEDS_SETUP:
            # The app is not installed. Don't try to set it up.
            return

        try:
            packages = list(
                itertools.chain.from_iterable(component.get_actual_packages()
                                              for component in components))
        except MissingPackageError:
            # If there are some packages needed by this app that are missing,
            # there is no way we can rerun setup for this app. Give up, don't
            # retry.
            return

        if not self._app_needs_setup_rerun(app, packages):
            # App does not need a setup rerun
            return

        operation = run_setup_on_app(app.app_id, rerun=True)
        if operation:
            operation.join()

    def _app_needs_setup_rerun(self, app: app_module.App,
                               packages: list[str]) -> bool:
        """Return whether an app needs an rerun."""
        packages_known = package.get_known()
        cache = apt.Cache()
        for package_ in packages:
            try:
                cache_package = cache[package_]
            except KeyError:
                logger.warning('For installed app %s, package %s is not known',
                               app.app_id, package_)
                return False

            if not cache_package.installed:
                # App is installed but one of the needed packages is not
                # installed. Don't know what to do. Don't rerun.
                logger.warning(
                    'For installed app %s, package %s is not installed',
                    app.app_id, package_)
                return False

            installed_version = cache_package.installed.version
            package_known = packages_known.get(package_, {})
            version_known = package_known.get('version')
            if installed_version != version_known:
                return True

        # Latest versions of all packages of the app have already been
        # processed (and thus known).
        return False

    def shutdown(self) -> None:
        """If we are sleeping for next attempt, cancel it.

        If we are actually performing operations, do nothing.
        """
        self._wait_event.set()


def on_dpkg_invoked():
    """Called by D-Bus service when dpkg has been invoked."""
    dpkg_handler = DpkgHandler.get_instance()
    dpkg_handler.on_dpkg_invoked()


def store_error_message(error_message: str):
    """Add an error message to thread local storage."""
    try:
        thread_local_storage.errors.append(error_message)
    except AttributeError:
        thread_local_storage.errors = [error_message]


def retrieve_error_messages() -> list[str]:
    """Retrieve the error messages from thread local storage.

    Errors are cleared after retrieval."""
    try:
        errors = thread_local_storage.errors
        thread_local_storage.errors = []
    except AttributeError:
        errors = []

    return errors



# SPDX-License-Identifier: AGPL-3.0-or-later
"""List and handle system services."""

from plinth import action_utils
from plinth import app as app_module
from plinth import module_loader
from plinth.actions import privileged
from plinth.daemon import Daemon, RelatedDaemon


@privileged
def systemd_set_default(target: str):
    """Set the default target that systemd will boot into."""
    if target not in ['graphical.target', 'multi-user.target']:
        raise ValueError('Invalid target')

    action_utils.systemd_set_default(target)


@privileged
def start(service: str):
    """Start a service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_start(service)


@privileged
def stop(service: str):
    """Stop a running service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_stop(service)


@privileged
def enable(service: str):
    """Enable a service so that it start on system boot."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_enable(service)


@privileged
def disable(service: str):
    """Disable a service so that it does not start on system boot."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_disable(service)


@privileged
def restart(service: str):
    """Restart a service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_restart(service)


@privileged
def try_restart(service: str):
    """Restart a service if it is running."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_try_restart(service)


@privileged
def reload(service: str):
    """Reload a service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_reload(service)


@privileged
def try_reload_or_restart(service: str):
    """Reload a service if it supports reloading, otherwise restart.

    Do nothing if service is not running.
    """
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_try_reload_or_restart(service)


@privileged
def mask(service: str):
    """Mask a service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_mask(service)


@privileged
def unmask(service: str):
    """Unmask a service."""
    _assert_service_is_managed_by_plinth(service)
    action_utils.service_unmask(service)


@privileged
def is_enabled(service: str) -> bool:
    """Return whether a service is enabled."""
    _assert_service_is_managed_by_plinth(service)
    return action_utils.service_is_enabled(service)


@privileged
def is_running(service: str) -> bool:
    """Return whether a service is running."""
    _assert_service_is_managed_by_plinth(service)
    return action_utils.service_is_running(service)


@privileged
def get_logs(service: str) -> dict[str, str]:
    _assert_service_is_managed_by_plinth(service)
    return {
        'unit': service,
        'description': action_utils.service_show(service)['Description'],
        'logs': action_utils.service_get_logs(service)
    }


def _get_managed_services():
    """Get a set of all services managed by FreedomBox."""
    from plinth.container import Container

    services = set()
    module_loader.load_modules()
    app_module.apps_init()
    for app in app_module.App.list():
        components = app.get_components_of_type(Daemon)
        for component in components:
            services.add(component.unit)
            if component.alias:
                services.add(component.alias)

        components = app.get_components_of_type(RelatedDaemon)
        for component in components:
            services.add(component.unit)

        components = app.get_components_of_type(Container)
        for component in components:
            services.add(component.name)

    return services


def _assert_service_is_managed_by_plinth(service_name):
    managed_services = _get_managed_services()
    if service_name not in managed_services:
        msg = ("The service '%s' is not managed by FreedomBox. Access is only "
               "permitted for services listed in the Daemon and RelatedDaemon "
               "components of any FreedomBox app.") % service_name
        raise ValueError(msg)



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

from . import privileged

_orbot_package_id = 'org.torproject.android'
_tor_browser_download_url = \
    'https://www.torproject.org/download/download-easy.html'

clients = [{
    'name':
        _('Tor Browser'),
    'platforms': [{
        'type': 'download',
        'os': 'windows',
        'url': _tor_browser_download_url,
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _tor_browser_download_url,
    }, {
        'type': 'download',
        'os': 'macos',
        'url': _tor_browser_download_url,
    }]
}, {
    'name':
        _('Orbot: Proxy with Tor'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _orbot_package_id)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _orbot_package_id)
    }]
}]

backup = {
    'config': {
        'directories': ['/etc/tor/instances/plinth/'],
        'files': [str(privileged.TOR_APACHE_SITE)]
    },
    'secrets': {
        'directories': ['/var/lib/tor-instances/plinth/']
    },
    'services': ['tor@plinth']
}

tags = [
    _('Onion services'),
    _('Relay'),
    _('Anonymity network'),
    _('Censorship resistance'),
    _('Tor'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Tests for help views.

Design: - Make tests independent from URL policy by using Django names instead
          of URLs to call the help module. For this, some additional fixture
          work is needed: pytestmark and fixture_app_urls().

Pending: - status log

"""

import json
import pathlib
import subprocess
from unittest.mock import Mock, patch

import pytest
from django import urls
from django.conf import settings
from django.http import Http404
from django.urls import re_path

from plinth import cfg, module_loader
from plinth.modules import help as help_module
from plinth.modules.help import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')


def _is_page(response):
    """Minimal check on help views."""
    return (response.status_code == 200 and 'title' in response.context_data
            and response['content-type'] == 'text/html; charset=utf-8')


@pytest.fixture(autouse=True, scope='module')
def fixture_app_urls():
    """Make sure app's URLs are part of plinth.urls."""
    urls = [
        re_path(r'^apps/$', views.index, name='apps'),
        re_path(r'^system/$', views.index, name='system')
    ]
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=urls):
        modules.append('plinth.modules.help')
        module_loader.include_urls()
        yield


@pytest.fixture(autouse=True, scope='module')
def fixture_developer_configuration():
    """Make sure docs are read from current directory instead of /usr."""
    cfg.read_file(cfg.get_develop_config_path())


@pytest.fixture(name='menu', autouse=True)
def fixture_menu(test_menu):
    """Initialized menu module."""
    help_module.HelpApp()  # Create all menu components


@pytest.mark.parametrize("view_name, view", (
    ('feedback', views.feedback),
    ('support', views.support),
    ('index', views.index),
))
def test_simple_help_pages(rf, view_name, view):
    """Simple common test for certain help views."""
    response = view(rf.get(urls.reverse('help:' + view_name)))
    assert _is_page(response)


@patch('apt.Cache')
@patch('gzip.decompress')
@patch('requests.get')
def test_contribute_page(requests_get, decompress, apt_cache, rf):
    """Test the contribute page."""
    issues = [{
        'type': 'testing-autorm',
        'packages': ['autormpkg'],
        'source': 'autormsrc',
        'bugs': ['123']
    }, {
        'type': 'no-testing',
        'package': ['notestingpkg'],
        'source': 'nottestingsrc',
    }, {
        'type': 'gift',
        'package': ['giftpkg'],
        'bug': '456',
        'title': 'bug 456 title',
    }, {
        'type': 'help',
        'package': ['helppkg'],
        'bug': '567',
        'title': 'bug 567 title',
    }]
    decompress.return_value = json.dumps(issues)
    response = views.contribute(rf.get(urls.reverse('help:contribute')))
    assert _is_page(response)
    assert issues == (response.context_data['testing_autorm'] +
                      response.context_data['no_testing'] +
                      response.context_data['gift'] +
                      response.context_data['help'])


@patch('plinth.modules.upgrades.views.is_newer_version_available')
@patch('plinth.modules.upgrades.views.get_os_release')
def test_about(_get_os_release, _is_newer_version_available, rf):
    """Test some expected items in about view."""
    about_url = urls.reverse('help:about')
    request = rf.get(about_url)
    request.user = Mock()
    request.user.is_authenticated = True
    response = views.about(request)
    assert _is_page(response)
    for item in ('version', 'new_version', 'os_release'):
        assert item in response.context_data

    request.user.is_authenticated = False
    response = views.about(request)
    for item in ('version', 'new_version', 'os_release'):
        assert item not in response.context_data


# ---------------------------------------------------------------------------
# Tests for serving the offline user guide ( the "manual")
#
# The manual can be requested:
# - Either complete on a single page or page by page.
# - Specifying (or not) the language.
# - The complete manual can be requested in HTML or PDF formats.
#
# Expected Behaviour Rules:
# - If the page isn't specified, the help module returns the full manual in
#   one single page.
# - The help module tries first to return the page in the specified
#   language. If not found (either that page doesn't exist or the language
#   wasn't secified) it falls back to its twin in the fallback language. If
#   it is neither available, it shows a proper error message.
#
# Design Decisions:
# - The PDF manual has a separate function to serve it.
# - The 'Manual' page doesn't exist as such. However there are files named
#   'Manual' containing the complete full manual in one single page.
# - In order to avoid loops, the fallback language is intercepted and
#   treated specifically.
#   - Problem: Requesting a missing page in a language that happens to be
#              the fallback one, missing it would cause the help module to
#              redirect to the same page, closing thereby a neverending loop.
#              The web served would probably break that loop, but it would
#              cause confusion to the user.
# - CI environments don't setup FreedomBox completely. A regular setup run is
#   impractically slow (10-15 mins), if even posible. Compiling and deploying
#   the manual is just 3 extra lines in .gitlab-ci.yml file:
#    - make -C doc
#    - mkdir -p /usr/share/freedombox/manual
#    - cp -r doc/manual /usr/share/freedombox/manual
#   But again, this causes the 4 minutes of test preparation to bump to 6.
#   It's not worth for just testing the offline manual, so the tests guess if
#   they are running in a restricted environment and skip.

canary = pathlib.Path('doc/manual/en/Coturn.part.html')
TRANSLATIONS = ('es', )
MANUAL_PAGES = ('Apache_userdir', 'APU', 'Backups', 'BananaPro', 'BeagleBone',
                'bepasty', 'Bind', 'Calibre', 'Cockpit', 'Configure',
                'Contribute', 'Coturn', 'Cubieboard2', 'Cubietruck',
                'DateTime', 'Debian', 'Deluge', 'Developer', 'Diagnostics',
                'Download', 'DynamicDNS', 'ejabberd', 'Firewall',
                'freedombox-manual', 'GettingHelp', 'GitWeb', 'Hardware',
                'Ikiwiki', 'Infinoted', 'Introduction', 'JSXC', 'LetsEncrypt',
                'Maker', 'MatrixSynapse', 'MediaWiki', 'Minetest', 'MiniDLNA',
                'Mumble', 'NameServices', 'Networks', 'OpenVPN',
                'OrangePiZero', 'PageKite', 'pcDuino3', 'Performance',
                'PineA64+', 'PioneerEdition', 'Plinth', 'Power', 'Privoxy',
                'Quassel', 'QuickStart', 'Radicale', 'RaspberryPi2',
                'RaspberryPi3B+', 'RaspberryPi3B', 'RaspberryPi4B',
                'ReleaseNotes', 'Rock64', 'RockPro64', 'Roundcube', 'Samba',
                'Searx', 'SecureShell', 'Security', 'ServiceDiscovery',
                'Shadowsocks', 'Sharing', 'Snapshots', 'Storage', 'Syncthing',
                'TinyTinyRSS', 'Tor', 'Transmission', 'Upgrades', 'USBWiFi',
                'Users', 'VirtualBox', 'WireGuard')
_restricted_reason = ('Needs installed manual. '
                      'CI speed-optimized workspace does not provide it.')
not_restricted_environment = pytest.mark.skipif(not canary.exists(),
                                                reason=_restricted_reason)


@pytest.mark.parametrize('lang', (None, '-'))
def test_full_default_manual(rf, lang):
    """Test request for the full default manual.

    Expected: Redirect to the full manual in the fallback language.

    """
    manual_url = urls.reverse('help:manual')
    response = views.manual(rf.get(manual_url), lang=lang)
    assert response.status_code == 302
    assert response.url == '/help/manual/en/'

    # With a language cookie set
    request = rf.get(manual_url)
    request.COOKIES[settings.LANGUAGE_COOKIE_NAME] = TRANSLATIONS[0]
    response = views.manual(request, lang=lang)
    assert response.status_code == 302
    assert response.url == f'/help/manual/{TRANSLATIONS[0]}/'


@pytest.mark.parametrize('lang', (None, '-'))
def test_default_manual_by_pages(rf, lang):
    """Test page-specific requests for the (default) manual.

    Expected: Redirect to their respective twins in the fallback language.
    Pending.: Redirect pages with plus-sign '+' in their name.

    """
    manual_url = urls.reverse('help:manual')
    for page in MANUAL_PAGES:
        if '+' in page or 'Manual' in page:  # Pine64+ & RaspberryPi3B+
            continue

        response = views.manual(rf.get(manual_url), lang=lang, page=page)
        assert response.status_code == 302
        assert response.url == '/help/manual/en/' + page

        # With a language cookie set
        request = rf.get(manual_url)
        request.COOKIES[settings.LANGUAGE_COOKIE_NAME] = TRANSLATIONS[0]
        response = views.manual(request, lang=lang, page=page)
        assert response.status_code == 302
        assert response.url == f'/help/manual/{TRANSLATIONS[0]}/{page}'


@not_restricted_environment
def test_specific_full_manual_translation(rf):
    """Test request for specific translated manuals.

    Expected: All return a page.

    """
    manual_url = urls.reverse('help:manual')
    for lang in ('es', 'en'):
        response = views.manual(rf.get(manual_url), lang=lang)
        assert _is_page(response)


@not_restricted_environment
def test_specific_manual_translation_by_pages(rf):
    """Test that translated-page-specific requests.

    Expected: All known page names return pages.

    """
    manual_url = urls.reverse('help:manual')
    for lang in ('es', 'en'):
        for page in MANUAL_PAGES:
            response = views.manual(rf.get(manual_url), page=page, lang=lang)
            assert _is_page(response)


@not_restricted_environment
def test_full_manual_requested_by_page_name(rf):
    """Test requests for 'Manual'.

    Note: 'Manual' is a file, not a manual page.
    Expected: Return a proper not found message (HTTP 404)
    Currently: Non fallback languages return a page.
               This is wrong, but doesn't cause any harm.

    """
    manual_url = urls.reverse('help:manual')
    page = 'Manual'

    for lang in TRANSLATIONS:
        response = views.manual(rf.get(manual_url), page=page, lang=lang)
        assert _is_page(response)

    with pytest.raises(Http404):
        views.manual(rf.get(manual_url), page=page, lang='en')


def test_missing_page(rf):
    """Test requests for missing pages.

    Expected:
    - Unspecified language: Fall back to its fallback twin.
    - Translated languages: Fall back to its fallback twin.
    - Fallback language...: Return a proper not found message (HTTP 404)
    - Unknown languages...: Fall back to its fallback twin.

    """
    manual_url = urls.reverse('help:manual')
    page = 'unknown'
    for lang in TRANSLATIONS + ('unknown', None):
        response = views.manual(rf.get(manual_url), page=page, lang=lang)
        assert response.status_code == 302
        assert response.url == '/help/manual/en/unknown'

    with pytest.raises(Http404):
        views.manual(rf.get(manual_url), page=page, lang='en')


@not_restricted_environment
def test_download_full_manual_file(rf, tmp_path):
    """Test download of manual.

    Design: - Downloads the default manual, a translated one and the
              fallback translation. None should fail. Then compares
              them.
            - Call diff command for fast comparision. Comparing the
              over 10MB bytestrings in python is insanely slow.

    """

    def _diff(file_name_a, file_name_b, same):
        file_a = tmp_path / file_name_a
        file_b = tmp_path / file_name_b
        process = subprocess.run(
            ['diff', '-q', str(file_a), str(file_b)], check=False)
        assert bool(process.returncode) != same

    url = urls.reverse('help:manual')
    manuals = {
        'unspecified': rf.get(url),
        'translated': rf.get(url, HTTP_ACCEPT_LANGUAGE='es'),
        'fallback': rf.get(url, HTTP_ACCEPT_LANGUAGE='en')
    }
    for name, request in manuals.items():
        response = views.download_manual(request)
        assert response.status_code == 200
        file = tmp_path / (name + '.pdf')
        file.write_bytes(response.content)

    _diff('fallback.pdf', 'unspecified.pdf', same=True)
    _diff('fallback.pdf', 'translated.pdf', same=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Provides privileged actions that run as root."""

from .aliases import setup_aliases
from .dkim import fix_incorrect_key_ownership, get_dkim_public_key, setup_dkim
from .domain import set_domains
from .home import setup_home
from .postfix import setup_postfix
from .spam import setup_spam

__all__ = [
    'setup_aliases', 'get_dkim_public_key', 'setup_dkim',
    'fix_incorrect_key_ownership', 'set_domains', 'setup_home',
    'setup_postfix', 'setup_spam'
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for snapshot.
"""

from django.utils.translation import gettext_lazy as _

backup = {
    'config': {
        'files': ['/etc/snapper/configs/root', '/etc/default/snapper']
    }
}

tags = [_('Periodic'), _('Restore'), _('Known good state'), _('Btrfs')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Common Django middleware.
"""

import logging
import re

from django import urls
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.db.utils import OperationalError
from django.http import Http404, HttpResponseNotAllowed
from django.shortcuts import redirect, render
from django.template.response import SimpleTemplateResponse
from django.utils.deprecation import MiddlewareMixin
from django.utils.translation import gettext as _
from stronghold.utils import is_view_func_public

from plinth import app as app_module
from plinth import setup
from plinth.utils import is_user_admin

from . import operation as operation_module
from . import views

logger = logging.getLogger(__name__)


def _collect_operations_results(request, app):
    """Show success/fail messages from previous operations."""
    operations = operation_module.manager.collect_results(app.app_id)
    for operation in operations:
        if operation.exception:
            views.messages_error(request, operation.translated_message,
                                 operation.exception)
        else:
            messages.success(request, operation.translated_message)


class SetupMiddleware(MiddlewareMixin):
    """Django middleware to show pre-setup message and setup progress."""

    @staticmethod
    def process_view(request, view_func, view_args, view_kwargs):
        """Handle a request as Django middleware request handler."""
        # Don't interfere with login page
        user_requests_login = request.path.startswith(
            urls.reverse(settings.LOGIN_URL))
        if user_requests_login:
            return

        # Perform a URL resolution. This is slightly inefficient as
        # Django will do this resolution again.
        try:
            resolver_match = urls.resolve(request.path_info)
        except urls.Resolver404:
            return

        non_app_namespaces = {'oauth2_provider'}
        if (not resolver_match.namespaces or not len(resolver_match.namespaces)
                or (set(resolver_match.namespaces) & non_app_namespaces)):
            # Requested URL does not belong to any application
            return

        app_id = resolver_match.namespaces[0]
        app = app_module.App.get(app_id)

        is_admin = is_user_admin(request)
        # Collect and show operations' results to admins
        if is_admin:
            _collect_operations_results(request, app)

        # Check if application is up-to-date
        if app.get_setup_state() == \
           app_module.App.SetupState.UP_TO_DATE:
            return

        if not is_admin:
            raise PermissionDenied

        # Only allow logged-in users to access any setup page
        view = login_required(views.SetupView.as_view())
        return view(request, app_id=app_id)


class AdminRequiredMiddleware(MiddlewareMixin):
    """Django middleware for authenticating requests for admin areas."""

    @staticmethod
    def check_user_group(view_func, request):
        if hasattr(view_func, 'GROUP_NAME'):
            return request.user.groups.filter(
                name=getattr(view_func, 'GROUP_NAME')).exists()

    @staticmethod
    def process_view(request, view_func, view_args, view_kwargs):
        """Reject non-admin access to views that are private and not marked."""
        if is_view_func_public(view_func) or \
           hasattr(view_func, 'IS_NON_ADMIN'):
            return

        public_urls = settings.STRONGHOLD_PUBLIC_URLS
        if any(re.match(url, request.path_info) for url in public_urls):
            return

        if is_user_admin(request):
            return

        if AdminRequiredMiddleware.check_user_group(view_func, request):
            return

        raise PermissionDenied


class FirstSetupMiddleware(MiddlewareMixin):
    """Django middleware to block all interactions before first setup."""

    @staticmethod
    def process_view(request, view_func, view_args, view_kwargs):
        """Block all user interactions when first setup is pending."""
        if not setup.is_first_setup_running:
            return

        context = {
            'is_first_setup_running': setup.is_first_setup_running,
            'refresh_page_sec': 3
        }
        return render(request, 'first_setup.html', context)


class CommonErrorMiddleware(MiddlewareMixin):
    """Django middleware to handle common errors."""

    @staticmethod
    def process_exception(request, exception):
        """Show a custom error page when OperationalError is raised."""
        logger.exception('Error processing page. %s %s, exception: %s',
                         request.method, request.path, exception)
        if isinstance(exception, OperationalError):
            message = _(
                'System is possibly under heavy load. Please retry later.')
            return SimpleTemplateResponse('error.html',
                                          context={'message': message},
                                          status=503)

        if isinstance(exception, Exception):
            match = request.resolver_match
            if not match.app_name and match.url_name == 'index':
                # Don't try to handle errors on the home page as it will lead
                # to infinite redirects.
                return None

            if isinstance(exception, Http404):
                message = _('Page not found: {url}').format(url=request.path)
                exception = None  # Don't show exception details
            elif request.method == 'POST':
                message = _('Error running operation.')
            else:
                message = _('Error loading page.')

            if exception:
                views.messages_error(request, message, exception)
            else:
                messages.error(request, message)

            redirect_url = CommonErrorMiddleware._get_redirect_url_on_error(
                request)
            return redirect(redirect_url)

        return None

    @staticmethod
    def process_response(request, response):
        """Handle 405 method not allowed errors.

        These errors may happen when we redirect to a page that does not allow
        GET.
        """
        if isinstance(response, HttpResponseNotAllowed):
            redirect_url = CommonErrorMiddleware._get_redirect_url_on_error(
                request)
            return redirect(redirect_url)

        return response

    @staticmethod
    def _get_redirect_url_on_error(request):
        """Return the URL to redirect to after an error."""
        if request.method != 'GET':
            return request.path

        # If the original request was a GET, trying to redirect to same URL
        # with same request method might result in an recursive loop. Instead
        # redirect to a parent URL.
        breadcrumbs = views.get_breadcrumbs(request)
        parent_index = 1 if len(breadcrumbs) > 1 else 0
        return list(breadcrumbs.keys())[parent_index]


class CSPDict(dict):
    """A dictionary to store Content Security Policy.

    And return a full value of the HTTP header.
    """

    def get_header_value(self) -> str:
        """Return the string header value for the policy stored."""
        return ' '.join([f'{key} {value};' for key, value in self.items()])


CONTENT_SECURITY_POLICY = CSPDict({
    # @fonts are allowed only from FreedomBox itself.
    'font-src': "'self'",
    # <frame>/<iframe> sources are disabled.
    'frame-src': "'none'",
    # <img> sources are allowed only from FreedomBox itself. Allow
    # data: URLs for SVGs in CSS.
    'img-src': "'self' data:",
    # Manifest file is not allowed as there is none yet.
    'manifest-src': "'none'",
    # <audio>, <video>, <track> tags are not allowed yet.
    'media-src': "'none'",
    # <object>, <embed>, <applet> tags are not allowed yet. No plugins
    # types are alllowed since object-src is 'none'.
    'object-src': "'none'",
    # Allow JS from FreedomBox itself (no inline and attribute
    # scripts).
    'script-src': "'self'",
    # Allow inline CSS and CSS files from Freedombox itself.
    'style-src': "'self' 'unsafe-inline'",
    # Web worker sources are allowed only from FreedomBox itself (for
    # JSXC).
    'worker-src': "'self'",
    # All other fetch sources including Ajax are not allowed from
    # FreedomBox itself.
    'default-src': "'self'",
    # <base> tag is not allowed.
    'base-uri': "'none'",
    # Enable strict sandboxing enabled with some exceptions:
    # - Allow running Javascript.
    # - Allow popups as sometimes we use <a target=_blank>
    # - Allow popups to have different sandbox requirements as we
    #   launch apps' web clients.
    # - Allow forms to support configuration forms.
    # - Allow policies to treat same origin differently from other
    # - origins
    # - Allow downloads such as backup tarballs.
    'sandbox': 'allow-scripts allow-popups '
               'allow-popups-to-escape-sandbox allow-forms '
               'allow-same-origin allow-downloads',
    # Form action should be to FreedomBox itself.
    'form-action': "'self'",
    # This interface may be not embedded in <frame>, <iframe>, etc.
    # tags.
    'frame-ancestors': "'none'",
})


class CommonHeadersMiddleware:

    def __init__(self, get_response):
        """Initialize the middleware object."""
        self.get_response = get_response

    def __call__(self, request):
        """Add common security middleware."""
        # Disable sending Referer (sic) header from FreedomBox web interface to
        # external websites. This improves privacy by not disclosing FreedomBox
        # domains/URLs to external domains. Apps such as blogs which want to
        # popularize themselves with referrer header may still do so.
        response = self.get_response(request)
        if not response.get('Referrer-Policy'):
            response['Referrer-Policy'] = 'same-origin'

        # Disable browser guessing of MIME types. FreedoBox already sets good
        # content types for all the common file types.
        if not response.get('X-Content-Type-Options'):
            response['X-Content-Type-Options'] = 'nosniff'

        csp = ' '.join([
            f'{key} {value};'
            for key, value in CONTENT_SECURITY_POLICY.items()
        ])
        if not response.get('Content-Security-Policy'):
            response['Content-Security-Policy'] = csp

        return response



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Symlink and unlink configuration files into /etc."""

import importlib
import inspect
import shutil

from plinth import app as app_module
from plinth import module_loader
from plinth.actions import privileged


def _get_managed_dropin_config(app_id: str, path: str):
    """Check that this is a path managed by the specified app."""
    module_path = module_loader.get_module_import_path(app_id)
    module = importlib.import_module(module_path)
    module_classes = inspect.getmembers(module, inspect.isclass)
    app_classes = [
        cls[1] for cls in module_classes if issubclass(cls[1], app_module.App)
    ]

    for cls in app_classes:
        app = cls()
        from plinth.config import DropinConfigs
        components = app.get_components_of_type(DropinConfigs)
        for component in components:
            if path in component.etc_paths:
                return component

    raise AssertionError('Not a managed drop-in config')


@privileged
def dropin_is_valid(app_id: str, path: str, copy_only: bool,
                    unlink_invalid: bool = False) -> bool:
    """Check if symlink from /etc/ to /usr/share/freedombox/etc is valid.

    Optionally, drop the link if it is invalid.
    """
    component = _get_managed_dropin_config(app_id, path)
    etc_path = component.get_etc_path(path)
    target = component.get_target_path(path)
    if etc_path.exists() or etc_path.is_symlink():
        if (not copy_only and etc_path.is_symlink()
                and etc_path.readlink() == target):
            return True

        if (copy_only and etc_path.is_file()
                and etc_path.read_text() == target.read_text()):
            return True

        if unlink_invalid:
            etc_path.unlink(missing_ok=True)

    return False


@privileged
def dropin_link(app_id: str, path: str, copy_only: bool):
    """Create a symlink from /etc/ to /usr/share/freedombox/etc."""
    component = _get_managed_dropin_config(app_id, path)
    target = component.get_target_path(path)
    etc_path = component.get_etc_path(path)
    etc_path.parent.mkdir(parents=True, exist_ok=True)
    if copy_only:
        shutil.copyfile(target, etc_path)
    else:
        etc_path.symlink_to(target)


@privileged
def dropin_unlink(app_id: str, path: str, missing_ok: bool = False):
    """Remove a symlink in /etc/."""
    component = _get_managed_dropin_config(app_id, path)
    etc_path = component.get_etc_path(path)
    etc_path.unlink(missing_ok=missing_ok)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for shaarli app.
"""

import time

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.shaarli]


class TestShaarliApp(functional.BaseAppTests):
    app_name = 'shaarli'
    has_service = False
    has_web = True

    def install_and_setup(self, session_browser):
        """Install the app and set it up if needed."""
        super().install_and_setup(session_browser)
        self._shaarli_is_setup(session_browser)

    def _shaarli_is_setup(self, session_browser):
        """Finish shaarli installation by creating user."""
        functional.access_url(session_browser, self.app_name)
        if not session_browser.url.endswith('/shaarli/install'):
            # Setup has already been completed.
            return

        username_field = session_browser.find_by_id('username')
        username_field.fill(functional.config['DEFAULT']['username'])

        password_field = session_browser.find_by_id('password')
        password_field.fill(functional.config['DEFAULT']['password'])

        button = session_browser.find_by_css(
            '#installform input[type=submit]')[0]
        functional.submit(session_browser, element=button)

    def _shaarli_login(self, session_browser):
        """Login to shaarli."""
        functional.access_url(session_browser, 'shaarli/login')

        login_form = session_browser.find_by_id('login-form')
        if not login_form:
            return

        login_field = login_form.find_by_name('login')
        login_field.fill(functional.config['DEFAULT']['username'])

        password_field = login_form.find_by_name('password')
        password_field.fill(functional.config['DEFAULT']['password'])

        login_form.find_by_css('input[type=submit]').click()

    def test_add_bookmark(self, session_browser):
        """Test adding a bookmark."""
        self._shaarli_login(session_browser)
        functional.access_url(session_browser, 'shaarli/admin/add-shaare')

        addlink_form = session_browser.find_by_id('addlink-form')
        addlink_form.find_by_id('shaare').fill('https://freedombox.org/')
        addlink_form.find_by_css('input[type=submit]').click()

        time.sleep(2)
        session_browser.find_by_id('button-save-edit').click()

        links = session_browser.links.find_by_href('https://freedombox.org/')
        assert len(links) > 0

    # TODO: Test deleting bookmarks.

    def test_uninstall(self, session_browser):
        """After uninstall test, after installing run initial setup."""
        super().test_uninstall(session_browser)
        self._shaarli_is_setup(session_browser)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Setup logging for the application.
"""

import logging
import logging.config
import sys
import typing
import warnings

from . import cfg

default_level = None


class LogEmitterProtocol(typing.Protocol):
    unit: str


class LogEmitter:
    """A mixin for App components that emit logs.

    Used as a simple base class for identifying components that have logs. Use
    the self.unit property to fetch systemd journal logs of the unit.
    """

    unit: str

    def get_logs(self: LogEmitterProtocol) -> dict[str, str]:
        from plinth.privileged import service as service_privileged
        return service_privileged.get_logs(self.unit)


class ColoredFormatter(logging.Formatter):
    """Print parts of log message in color."""
    codes = {
        'black': 30,
        'red': 31,
        'green': 32,
        'yellow': 33,
        'blue': 34,
        'magenta': 35,
        'cyan': 36,
        'white': 37,
        'bright_black': 90,
        'bright_red': 91,
        'bright_green': 92,
        'bright_yellow': 93,
        'bright_blue': 94,
        'bright_magenta': 95,
        'bright_cyan': 96,
        'bright_white': 97
    }

    level_colors = {
        'DEBUG': 'bright_black',
        'INFO': 'bright_white',
        'WARNING': 'bright_yellow',
        'ERROR': 'red',
        'CRITICAL': 'bright_red'
    }

    def wrap_color(self, string, color=None):
        """Return a string wrapped in terminal escape codes for coloring."""
        if not color:
            return string

        return '\x1b[{}m'.format(self.codes[color]) + string + '\x1b[0m'

    def format(self, record):
        """Format a record into a string"""
        record_name = '{:<20}'.format(record.name)
        record.colored_name = self.wrap_color(record_name, 'bright_blue')

        level_color = self.level_colors.get(record.levelname, None)
        level_name = '{:>8}'.format(record.levelname)
        record.colored_levelname = self.wrap_color(level_name, level_color)
        return super().format(record)


def _capture_warnings():
    """Capture all warnings include deprecation warnings."""
    # Capture all Python warnings such as deprecation warnings
    logging.captureWarnings(True)

    # Log all deprecation warnings when in develop mode
    if cfg.develop:
        warnings.filterwarnings('default', '', DeprecationWarning)
        warnings.filterwarnings('default', '', PendingDeprecationWarning)
        warnings.filterwarnings('default', '', ImportWarning)


def action_init(console: bool = False):
    """Initialize logging for action scripts."""
    _capture_warnings()

    configuration = get_configuration()
    if console:
        configuration['root']['handlers'] = ['console']
    else:
        configuration['root']['handlers'] = ['journal']

    logging.config.dictConfig(configuration)


def init():
    """Setup the logging framework."""
    import cherrypy

    # Remove default handlers and let the log message propagate to root logger.
    for cherrypy_logger in [cherrypy.log.error_log, cherrypy.log.access_log]:
        for handler in list(cherrypy_logger.handlers):
            cherrypy_logger.removeHandler(handler)

    _capture_warnings()


def setup_cherrypy_static_directory(app):
    """Hush output from cherrypy static file request logging.

    Static file serving logs are hardly useful.
    """
    app.log.access_log.propagate = False
    app.log.error_log.propagate = False


def get_configuration():
    """Return the main python logging module configuration."""
    configuration = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'color': {
                '()': 'plinth.log.ColoredFormatter',
                'format': '{colored_levelname} {colored_name} {message}',
                'style': '{'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'color'
            },
            'journal': {
                'class': 'systemd.journal.JournalHandler'
            }
        },
        'root': {
            'handlers': ['console', 'journal'],
            'level': default_level or ('DEBUG' if cfg.develop else 'INFO')
        },
        'loggers': {
            'django.db.backends': {
                'level': 'INFO'  # Set to 'DEBUG' to log database queries
            },
            'axes': {
                'level': 'INFO'  # Too verbose during DEBUG
            }
        }
    }
    if not sys.stdin.isatty():
        # If running on a console, say due to user manually starting it, then
        # log everything to console. However, if we are running as a systemd
        # unit then no need to log to console. This allows us to capture
        # stdout/stderr in systemd unit without repeating the log messages.
        configuration['root']['handlers'].remove('console')

    return configuration



# SPDX-License-Identifier: AGPL-3.0-or-later
"""The main method for a daemon that runs privileged methods."""

import argparse
import io
import json
import logging
import os
import pathlib
import pwd
import signal
import socket
import socketserver
import struct
import sys
import threading
import time

import systemd.daemon

from . import __version__, actions
from . import app as app_module
from . import log, module_loader

logger = logging.getLogger(__name__)

address = pathlib.Path('/run/freedombox/privileged.socket')

FREEDOMBOX_PROCESS_USER = 'plinth'

MAX_REQUEST_LENGTH = 1_000_000

_server = None

idle_shutdown_time: int | None = 5 * 60  # 5 minutes

freedombox_develop = False

EXIT_SYNTAX = 10
EXIT_PERM = 20


class RequestHandler(socketserver.StreamRequestHandler):
    """Handle a single streaming request.

    It is instantiated once per connection to the server. The overridden
    handle() method implements communication with the newly connected client.
    """

    def _read_request(self) -> str:
        """Return a single request read from the client."""
        request_data = self.rfile.read(MAX_REQUEST_LENGTH + 1)
        if len(request_data) > MAX_REQUEST_LENGTH:
            raise ValueError('Request too large')

        try:
            request = request_data.decode('utf-8')
        except UnicodeError:
            raise ValueError('Invalid Unicode in request')

        return request

    def _write_response(self, response: str | io.BufferedReader):
        """Write a single response to the client."""
        if isinstance(response, str):
            self.wfile.write(response.encode('utf-8'))
        else:
            for chunk in response:
                self.wfile.write(chunk)

    def handle(self) -> None:
        """Handle a new connection from a client."""
        try:
            request = self._read_request()
            response_string = actions.privileged_handle_json_request(request)
        except Exception as exception:
            logger.exception('Error running privileged request: %s', exception)
            response = actions.get_return_value_from_exception(exception)
            response_string = json.dumps(response)

        self._write_response(response_string)


class Server(socketserver.ThreadingUnixStreamServer):
    """Server to handle privileged request.

    Requests from any process with UID other than from FreedomBox daemon user
    or root will be denied.

    If the server does not receive a request for idle_shutdown_time seconds and
    no requests are being processed, then serve_forever() will raise
    TimeoutError (so that the program can catch it and exit).

    If the daemon is spawned by systemd socket activation, then systemd
    provided socket is re-used (no bind() and listen() calls are made on it)
    and it will not be closed after the server is shutdown. If a daemon is
    spawned without socket activation, unix socket file is created with
    appropriate permissions after the parent directory is created. socket
    object is created, bind() and listen() calls will be made on the socket.
    When server is shutdown, close() is called on the socket.
    """

    def __init__(self, address, *args, **kwargs):
        """Initialize the server."""
        # Retrieve FreedomBox process UID
        user_info = pwd.getpwnam(FREEDOMBOX_PROCESS_USER)
        self.allowed_peer_uids = [0, user_info.pw_uid]

        # Used to auto-shutdown service
        self.last_request_time = time.time()

        self.listen_fd = self._get_listen_fd()
        if self.listen_fd:
            logger.info('systemd socket activated.')
            self.socket = socket.fromfd(self.listen_fd, socket.AF_UNIX,
                                        socket.SOCK_STREAM)
            self.server_address = self.socket.getsockname()
            super(socketserver.TCPServer,
                  self).__init__(address, *args, **kwargs)
        else:
            logger.info('systemd service activated.')
            address_path = pathlib.Path(address)
            address_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
            address_path.unlink(missing_ok=True)
            super().__init__(address, *args, **kwargs)
            address_path.chmod(0o666)  # All users are allowed to connect

    def server_bind(self):
        """Called by constructor to activate the server.

        Do nothing if socket-activated by systemd.
        """
        if not self.listen_fd:
            super().server_bind()

    def server_activate(self):
        """Called by constructor to activate the server.

        Do nothing if socket-activated by systemd.
        """
        if not self.listen_fd:
            super().server_activate()

    def server_close(self):
        """Called to clean-up the server.

        Don't close the socket if socket-activated by systemd.
        """
        if not self.listen_fd:
            # Not called with systemd socket activation. We are responsible for
            # cleaning up the socket file we created.
            super().server_close()
            pathlib.Path(str(self.server_address)).unlink(missing_ok=True)
        else:
            # Don't close the socket, systemd is still using it for next
            # invocation.
            pass

    def service_actions(self):
        """Called from serve_forever() loop. Shutdown service if unused."""
        super().service_actions()

        if idle_shutdown_time is None:
            return

        if time.time() - self.last_request_time < idle_shutdown_time:
            return

        if not isinstance(self._threads, list):
            return

        if any(thread.is_alive() for thread in self._threads):
            return

        # Raise an exception in the serve_forever() loop.
        raise TimeoutError()

    def _get_listen_fd(self):
        """Return the listening socket from systemd socket activation."""
        listen_fds = systemd.daemon.listen_fds(unset_environment=True)
        if len(listen_fds) == 0:
            # Activated without socket activation.
            return None

        if len(listen_fds) > 1:
            # Activated with multiple listening sockets. We didn't configure
            # our .socket unit like this. This is unexpected.
            return None

        listen_fd = listen_fds[0]
        if not systemd.daemon.is_socket_unix(listen_fd, socket.SOCK_STREAM,
                                             listening=1):
            # Socket is not a AF_UNIX socket, it is not SOCK_STREAM socket, or
            # listen() has not been called on it. This is unexpected.
            return None

        return listen_fd

    def verify_request(self, request, client_address) -> bool:
        """Return False if the request must be denied."""
        creds = request.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED,
                                   struct.calcsize('3i'))
        _, uid, _ = struct.unpack('3i', creds)
        if uid not in self.allowed_peer_uids:
            return False

        self.last_request_time = time.time()

        return True


def client_main() -> None:
    """Parse arguments for the client for privileged daemon."""
    log.action_init(console=True)

    parser = argparse.ArgumentParser()
    parser.add_argument('module', help='Module to trigger action in')
    parser.add_argument('action', help='Action to trigger in module')
    parser.add_argument('--no-args', default=False, action='store_true',
                        help='Do not read arguments from stdin')
    args = parser.parse_args()

    try:
        try:
            arguments: dict = {'args': [], 'kwargs': {}}
            if not args.no_args:
                input_ = sys.stdin.read()
                if input_:
                    arguments = json.loads(input_)
        except json.JSONDecodeError as exception:
            raise SyntaxError('Arguments on stdin not JSON.') from exception

        return_value = actions.run_privileged_method(None, args.module,
                                                     args.action,
                                                     arguments['args'],
                                                     arguments['kwargs'])
        print(json.dumps(return_value, cls=actions.JSONEncoder))
    except PermissionError as exception:
        logger.error(exception.args[0])
        sys.exit(EXIT_PERM)
    except SyntaxError as exception:
        logger.error(exception.args[0])
        sys.exit(EXIT_SYNTAX)
    except TypeError as exception:
        logger.error(exception.args[0])
        sys.exit(EXIT_SYNTAX)
    except Exception as exception:
        logger.exception(exception)
        sys.exit(1)


def _on_sigterm(signal_number: int, frame) -> None:
    """Handle SIGTERM signal. Issue server shutdown."""
    threading.Thread(target=_shutdown_server).start()


def _shutdown_server() -> None:
    """Issue a shutdown request to the server.

    This must be run in a thread separate from the server.serve_forever()
    otherwise it will deadlock waiting for the shutdown to complete.
    """
    logger.info('SIGTERM received, shutting down the server.')
    if _server:
        _server.shutdown()

    logger.info('Shutdown complete, some requests may be running.')


def main() -> None:
    """Start the server, listen on socket, and serve forever."""
    global freedombox_develop, idle_shutdown_time

    log.action_init()

    logger.info('FreedomBox privileged daemon: %s', __version__)

    if os.getenv('FREEDOMBOX_DEVELOP', '') == '1':
        freedombox_develop = True
        idle_shutdown_time = 5
        logger.info('Running development mode, idle shutdown time = %ss',
                    idle_shutdown_time)

    # When invoked as a systemd service, don't perform automatic idle shutdown.
    if not systemd.daemon.listen_fds(unset_environment=False):
        idle_shutdown_time = None

    signal.signal(signal.SIGTERM, _on_sigterm)

    module_loader.load_modules()
    app_module.apps_init()

    with Server(str(address), RequestHandler) as server:
        global _server
        _server = server  # Reference needed to shutdown the server.

        # systemd will wait until notification to proceed with other processes.
        # We have service Type=notify.
        systemd.daemon.notify('READY=1')

        try:
            server.serve_forever()
        except TimeoutError:
            logger.info('FreedomBox privileged daemon exiting on idle.')
        except Exception as exception:
            logger.exception(
                'FreedomBox privileged daemon exiting on error - %s',
                exception)
            sys.exit(-1)
        else:
            logger.info('FreedomBox privileged daemon exiting.')

        # Exit the context manager. This calls server.close() which waits on
        # all pending request threads to complete.

    logger.info('All requested completed. Exit.')


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url
from plinth.modules.jsxc import manifest as jsxc_manifest

_clients = [{
    'name':
        _('Conversations'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'eu.siacs.conversations')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'eu.siacs.conversations')
    }]
}, {
    'name':
        _('Xabber'),
    'description':
        _('Open source Jabber (XMPP) client with multi-account support '
          'and clean and simple interface. '),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'com.xabber.androiddev')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'com.xabber.android')
    }]
}, {
    'name':
        _('Yaxim'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'org.yaxim.androidclient'),
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'org.yaxim.androidclient'),
    }]
}, {
    'name':
        _('Monal - XMPP Chat'),
    'platforms': [{
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/us/app/monal-xmpp-chat/id317711500'
    }]
}, {
    'name':
        _('Siskin IM'),
    'platforms': [{
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/us/app/tigase-messenger/id1153516838'
    }]
}, {
    'name':
        _('Dino'),
    'platforms': [{
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://github.com/dino/dino/wiki/Distribution-Packages',
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'dino-im'
    }]
}, {
    'name':
        _('Gajim'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'gajim'
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://gajim.org/downloads.php'
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://gajim.org/downloads.php'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://gajim.org/downloads.php'
    }]
}]

_clients.extend(jsxc_manifest.clients)

clients = _clients

backup = {
    'config': {
        'files': ['/etc/ejabberd/ejabberd.yml']
    },
    'data': {
        'directories': ['/var/lib/ejabberd/']
    },
    'secrets': {
        'files': ['/etc/ejabberd/ejabberd.pem'],
        'directories': ['/etc/ejabberd/letsencrypt/']
    },
    'services': ['ejabberd']
}

tags = [
    _('Encrypted messaging'),
    _('Audio chat'),
    _('Video chat'),
    _('XMPP'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Handle container run using podman."""

from plinth import action_utils
from plinth import app as app_module
from plinth import module_loader
from plinth.actions import privileged


@privileged
def container_is_enabled(container: str) -> bool:
    """Return whether a container is enabled."""
    _assert_container_is_managed(container)
    return action_utils.podman_is_enabled(container)


@privileged
def container_enable(container: str):
    """Enable a container so that it start on system boot."""
    _assert_container_is_managed(container)
    action_utils.podman_enable(container)
    action_utils.service_enable(container)


@privileged
def container_disable(container: str):
    """Disable a container so that it does not start on system boot."""
    _assert_container_is_managed(container)
    action_utils.service_disable(container)
    action_utils.podman_disable(container)


@privileged
def container_setup(container: str, image_name: str, volume_name: str,
                    volume_path: str, volumes: dict[str, str] | None = None,
                    env: dict[str, str] | None = None,
                    binds_to: list[str] | None = None,
                    devices: dict[str, str] | None = None):
    """Remove and recreate the podman container."""
    _assert_container_is_managed(container)
    action_utils.podman_create(container, image_name, volume_name, volume_path,
                               volumes, env, binds_to, devices)
    action_utils.service_start(container, check=True)


@privileged
def container_uninstall(container: str, image_name: str, volume_name: str,
                        volume_path: str):
    """Remove podman container."""
    action_utils.podman_uninstall(container_name=container,
                                  image_name=image_name,
                                  volume_name=volume_name,
                                  volume_path=volume_path)


def _get_managed_containers() -> set[str]:
    """Get a set of all containers managed by FreedomBox."""
    from plinth.container import Container

    containers = set()
    module_loader.load_modules()
    app_module.apps_init()
    for app in app_module.App.list():
        components = app.get_components_of_type(Container)
        for component in components:
            containers.add(component.name)

    return containers


def _assert_container_is_managed(container_name):
    """Check that container is managed by one of the FreedomBox apps."""
    managed_containers = _get_managed_containers()
    if container_name not in managed_containers:
        msg = ("The container '%s' is not managed by FreedomBox. Access is "
               "only permitted for containers listed in the Container "
               "components of any FreedomBox app.") % container_name
        raise ValueError(msg)
